# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides definitions and untilities for connecting to the Additive server."""
from __future__ import annotations

from dataclasses import dataclass
import logging
import time

from ansys.api.additive.v0.about_pb2_grpc import AboutServiceStub
from ansys.api.additive.v0.additive_materials_pb2_grpc import MaterialsServiceStub
from ansys.api.additive.v0.additive_simulation_pb2_grpc import SimulationServiceStub
import ansys.platform.instancemanagement as pypim
from google.protobuf.empty_pb2 import Empty
import grpc

from ansys.additive.core.server_connection.constants import (
    DEFAULT_PRODUCT_VERSION,
    LOCALHOST,
    PYPIM_PRODUCT_NAME,
)
from ansys.additive.core.server_connection.local_server import LocalServer
from ansys.additive.core.server_connection.network_utils import create_channel


@dataclass(frozen=True)
class ServerConnectionStatus:
    """Provides status information about a server.

    Parameters
    ----------
    connected: bool
        True if server is connected.
    channel_str: string, None
        Hostname and port of server connection in the form ``host:port``.
    metadata: dict, None
        Server metadata.
    """

    connected: bool
    """True if server is connected."""
    channel_str: str = None
    """Hostname and port of server connection in the form ``host:port``."""
    metadata: dict = None
    """Server metadata."""


class ServerConnection:
    """Provides connection to Additive server.

    If neither ``channel`` nor ``addr`` are provided, an attempt is
    made to start an Additive server and connect to it. Starting a server
    in a cloud environment requires
    `PyPIM <https://pypim.docs.pyansys.com/version/stable/index.html>`_ to be available.
    To start a server when running on localhost, the Additive option of the
    Structures package of the Ansys unified installation must be installed.

    Parameters
    ----------
    channel: grpc.Channel, None
        gRPC channel connected to server.
    addr: str, None
        IPv4 address of server of the form ``host:port``.
    product_version: str
        Version of the Ansys product installation in the form ``"YYR"``, where ``YY``
        is the two-digit year and ``R`` is the release number. For example, the release
        2024 R1 would be specified as ``241``. This parameter is only applicable in
        `PyPIM <https://pypim.docs.pyansys.com/version/stable/index.html>`_-enabled
        cloud environments and on localhost.
    log: logging.Logger, None
        Log to write connection messages to.
    """

    def __init__(
        self,
        channel: grpc.Channel | None = None,
        addr: str | None = None,
        product_version: str = DEFAULT_PRODUCT_VERSION,
        log: logging.Logger = None,
    ) -> None:
        """Initialize a server connection."""

        if channel is not None and addr is not None:
            raise ValueError("Both 'channel' and 'addr' cannot both be specified.")

        self._log = log if log else logging.getLogger(__name__)

        if channel:
            self._channel = channel
        else:
            if addr:
                target = addr
            elif pypim.is_configured():
                pim = pypim.connect()
                self._server_instance = pim.create_instance(
                    product_name=PYPIM_PRODUCT_NAME, product_version=product_version
                )
                self._log.info("Waiting for server to initialize")
                self._server_instance.wait_for_ready()
                (_, target) = self._server_instance.services["grpc"].uri.split(":", 1)
            else:
                port = LocalServer.find_open_port()
                self._server_process = LocalServer.launch(port, product_version=product_version)
                target = f"{LOCALHOST}:{port}"
            self._channel = create_channel(target)

        # assign service stubs
        self._materials_stub = MaterialsServiceStub(self._channel)
        self._simulation_stub = SimulationServiceStub(self._channel)
        self._about_stub = AboutServiceStub(self._channel)

        if not self.ready():
            raise RuntimeError(f"Unable to connect to server {self.channel_str}")

        self._log.info("Connected to %s", self.channel_str)

    def __del__(self):
        """Destructor for cleaning up server connection."""
        if hasattr(self, "_server_instance") and self._server_instance:
            self._server_instance.delete()
        if hasattr(self, "_server_process") and self._server_process:
            self._server_process.kill()

    @property
    def channel_str(self) -> str:
        """GRPC channel target.

        The form is generally ``"ip:port"``. For example, ``"127.0.0.1:50052"``.
        """
        if self._channel is not None:
            return self._channel._channel.target().decode()
        return ""

    @property
    def materials_stub(self) -> MaterialsServiceStub:
        """Materials service stub."""
        return self._materials_stub

    @property
    def simulation_stub(self) -> SimulationServiceStub:
        """Simulation service stub."""
        return self._simulation_stub

    def status(self) -> ServerConnectionStatus:
        """Return the server connection status."""
        if not hasattr(self, "_channel") or self._channel is None:
            return ServerConnectionStatus(False)
        try:
            response = self._about_stub.About(Empty())
        except grpc.RpcError:
            return ServerConnectionStatus(False, self.channel_str)
        metadata = {}
        for key in response.metadata:
            metadata[key] = response.metadata[key]
        return ServerConnectionStatus(True, self.channel_str, metadata)

    def ready(self, retries: int = 5) -> bool:
        """Return whether the server is ready.

        Parameters
        ----------
        retries: int
            Number of times to retry before giving up. An linearly increasing delay
            is used between each retry.

        Returns
        -------
        bool:
            True means server is ready. False means the number of retries was exceeded
            without receiving a response from the server.
        """
        ready = False
        for i in range(retries + 1):
            try:
                self._about_stub.About(Empty())
                ready = True
                break
            except grpc.RpcError:
                time.sleep(i + 1)

        return ready



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a contour plot for relative density and build rate."""
from __future__ import annotations

import os

import numpy as np
import pandas as pd
import panel as pn
import plotly.graph_objects as go

from ansys.additive.core import SimulationStatus, SimulationType
from ansys.additive.core.misc import short_uuid
from ansys.additive.core.parametric_study import ColumnNames, ParametricStudy

from ._common_controls import _common_controls

# Initialize panel for plotly.
pn.extension("plotly")


def porosity_contour_plot(ps: ParametricStudy):
    """Generates a contour plot of build rate and relative density.

    Parameters
    ----------
    ps : ParametricStudy
        Parametric study to plot.

    Returns
    -------
    panel.Row
        Interactive plot.
    """
    df = __data_frame(ps)
    (
        ht_select,
        lt_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
        show_scatter_cb,
        show_contours_cb,
    ) = __init_controls(df)
    row1 = pn.Column(
        show_scatter_cb,
        show_contours_cb,
        lt_select,
        ht_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
        width=200,
    )
    plot_view = pn.bind(
        __update_plot,
        df,
        ht_select,
        lt_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
        show_scatter_cb,
        show_contours_cb,
    )
    plot = pn.Row(
        row1,
        pn.pane.Plotly(plot_view, sizing_mode="stretch_both", min_height=600),
        sizing_mode="stretch_both",
    ).servable()
    if os.getenv("GENERATING_DOCS"):
        name = porosity_contour_plot.__name__
        plot.save(f"{name}_{short_uuid()}.png")
        plot.__repr__ = lambda: name
    return plot


def __data_frame(ps: ParametricStudy) -> pd.DataFrame:
    df = ps.data_frame()
    df = df[
        (df[ColumnNames.TYPE] == SimulationType.POROSITY)
        & (df[ColumnNames.STATUS] == SimulationStatus.COMPLETED)
    ]
    if len(df.index) < 2:
        raise ValueError("Too few data points to plot")
    # convert build rate from m^3/s to mm^3/s
    df.loc[:, ColumnNames.BUILD_RATE] *= 1e9
    return df


def __init_controls(df: pd.DataFrame):
    (
        ht_select,
        lt_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
    ) = _common_controls(df)
    show_scatter_cb = pn.widgets.Checkbox(
        name="Relative Density Points", sizing_mode="stretch_width"
    ).servable()
    show_scatter_cb.value = True
    show_contours_cb = pn.widgets.Checkbox(
        name="Relative Density Contours", sizing_mode="stretch_width"
    ).servable()
    show_contours_cb.value = True
    return (
        ht_select,
        lt_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
        show_scatter_cb,
        show_contours_cb,
    )


def __update_plot(
    df: pd.DataFrame,
    ht: float,
    lt: float,
    bd: float,
    sa: float,
    ra: float,
    hs: float,
    sw: float,
    show_scatter: bool,
    show_contours: bool,
) -> dict:
    fig = go.Figure()

    x, y, br, rd = __contour_data(df, ht, lt, bd, sa, ra, hs, sw)
    br_contour = go.Contour(
        x=x,
        y=y,
        z=br,
        contours=dict(showlabels=True, labelfont=dict(size=12, color="black")),
        contours_coloring="lines",
        line=dict(dash="dot", width=2),
        colorscale="Phase",
        colorbar=dict(
            title="Build Rate (mm^3/s)",
            titlefont=dict(size=12),
            titleside="right",
            tickfont=dict(size=12),
            thickness=20,
            len=0.25,
            lenmode="fraction",
            y=0.2,
        ),
        connectgaps=True,
    )
    fig.add_trace(br_contour)

    rd_contour = go.Contour(
        x=x,
        y=y,
        z=rd,
        contours=dict(showlabels=True, labelfont=dict(size=12, color="black")),
        contours_coloring="lines",
        line=dict(dash="solid", width=2),
        colorscale="Rainbow",
        colorbar=dict(
            title="Relative Density",
            titlefont=dict(size=12),
            titleside="right",
            thickness=20,
            tickfont=dict(size=12),
            len=0.25,
            lenmode="fraction",
            y=0.6,
        ),
        connectgaps=True,
        visible=show_contours,
    )
    fig.add_trace(rd_contour)

    scatter_x, scatter_y, rd_scatter = __scatter_data(df, ht, lt, bd, sa, ra, hs, sw)
    scatter = go.Scatter(
        x=scatter_x,
        y=scatter_y,
        mode="markers+text",
        text=[f"{z:.4f}" for z in rd_scatter],
        textposition="top center",
        visible=show_scatter,
        marker=dict(color="slategrey", size=5),
    )
    fig.add_trace(scatter)
    fig.update_layout(
        title=dict(
            text="Relative Density and Build Rate"
        )  # , xanchor="center", yanchor="top", x=0.5, y=0.9)
    )
    fig.update_xaxes(
        title_text="Scan Speed (m/s)",
        showgrid=True,
        gridwidth=1,
        gridcolor="white",
        title_font=dict(size=12),
        range=[min(x) - 0.1, max(x) + 0.1],
    )
    fig.update_yaxes(
        title_text="Laser Power (W)",
        showgrid=True,
        gridwidth=1,
        gridcolor="white",
        title_font=dict(size=12),
        range=[min(y) - 20, max(y) + 70],
    )

    return fig


def __contour_data(
    df: pd.DataFrame, ht: float, lt: float, bd: float, sa: float, ra: float, hs: float, sw: float
) -> tuple[list, list, list, list]:
    """Returns lists of scan speed, laser power, build rate, and relative
    density values."""

    idx = df[
        (df[ColumnNames.LAYER_THICKNESS] == lt)
        & (df[ColumnNames.BEAM_DIAMETER] == bd)
        & (df[ColumnNames.HEATER_TEMPERATURE] == ht)
        & (df[ColumnNames.START_ANGLE] == sa)
        & (df[ColumnNames.ROTATION_ANGLE] == ra)
        & (df[ColumnNames.HATCH_SPACING] == hs)
        & (df[ColumnNames.STRIPE_WIDTH] == sw)
    ].index

    df = df.loc[
        idx,
        [
            ColumnNames.LASER_POWER,
            ColumnNames.SCAN_SPEED,
            ColumnNames.BUILD_RATE,
            ColumnNames.RELATIVE_DENSITY,
        ],
    ]
    df.sort_values(
        by=[ColumnNames.LASER_POWER, ColumnNames.SCAN_SPEED],
        inplace=True,
    )
    speeds = df[ColumnNames.SCAN_SPEED].unique()
    powers = df[ColumnNames.LASER_POWER].unique()
    build_rate_z = []
    relative_density_z = []
    for p in powers:
        br_row = []
        rd_row = []
        for v in speeds:
            if (
                len(
                    df.loc[
                        (df[ColumnNames.LASER_POWER] == p) & (df[ColumnNames.SCAN_SPEED] == v),
                        ColumnNames.BUILD_RATE,
                    ].values
                )
                > 0
            ):
                br_row.append(
                    df.loc[
                        (df[ColumnNames.LASER_POWER] == p) & (df[ColumnNames.SCAN_SPEED] == v),
                        ColumnNames.BUILD_RATE,
                    ].values[0]
                )
            else:
                br_row.append(np.nan)
            if (
                len(
                    df.loc[
                        (df[ColumnNames.LASER_POWER] == p) & (df[ColumnNames.SCAN_SPEED] == v),
                        ColumnNames.RELATIVE_DENSITY,
                    ].values
                )
                > 0
            ):
                rd_row.append(
                    df.loc[
                        (df[ColumnNames.LASER_POWER] == p) & (df[ColumnNames.SCAN_SPEED] == v),
                        ColumnNames.RELATIVE_DENSITY,
                    ].values[0]
                )
            else:
                rd_row.append(np.nan)
        build_rate_z.append(br_row)
        relative_density_z.append(rd_row)
    return (speeds, powers, build_rate_z, relative_density_z)


def __scatter_data(
    df: pd.DataFrame, ht: float, lt: float, bd: float, sa: float, ra: float, hs: float, sw: float
) -> tuple[list, list, list]:
    idx = df[
        (df[ColumnNames.LAYER_THICKNESS] == lt)
        & (df[ColumnNames.BEAM_DIAMETER] == bd)
        & (df[ColumnNames.HEATER_TEMPERATURE] == ht)
        & (df[ColumnNames.START_ANGLE] == sa)
        & (df[ColumnNames.ROTATION_ANGLE] == ra)
        & (df[ColumnNames.HATCH_SPACING] == hs)
        & (df[ColumnNames.STRIPE_WIDTH] == sw)
    ].index

    df = df.loc[
        idx,
        [
            ColumnNames.LASER_POWER,
            ColumnNames.SCAN_SPEED,
            ColumnNames.BUILD_RATE,
            ColumnNames.RELATIVE_DENSITY,
        ],
    ]
    df.sort_values(
        by=[ColumnNames.LASER_POWER, ColumnNames.SCAN_SPEED],
        inplace=True,
    )
    return (
        df[ColumnNames.SCAN_SPEED].tolist(),
        df[ColumnNames.LASER_POWER].tolist(),
        df[ColumnNames.RELATIVE_DENSITY].tolist(),
    )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides input and result summary containers for microstructure simulations."""
import math
import os

from ansys.api.additive.v0.additive_domain_pb2 import (
    MicrostructureInput as MicrostructureInputMessage,
)
from ansys.api.additive.v0.additive_domain_pb2 import (
    MicrostructureResult as MicrostructureResultMessage,
)
from ansys.api.additive.v0.additive_simulation_pb2 import SimulationRequest
from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
import numpy as np
import pandas as pd

from ansys.additive.core import misc
from ansys.additive.core.machine import AdditiveMachine
from ansys.additive.core.material import AdditiveMaterial


class MicrostructureInput:
    """Provides input parameters for microstructure simulation.

    Units are SI (m, kg, s, K) unless otherwise noted.
    """

    DEFAULT_POSITION_COORDINATE = 0
    """Default X, Y, Z, position coordinate (m)."""
    MIN_POSITION_COORDINATE = 0
    """Minimum X, Y, Z, position coordinate (m)."""
    MAX_POSITION_COORDINATE = 10
    """Maximum X, Y, Z, position coordinate (m)."""
    DEFAULT_SAMPLE_SIZE = 1.5e-3
    """Default sample size in each dimension (m)."""
    MIN_SAMPLE_SIZE = 0.001
    """Minimum sample size in each dimension (m)."""
    MAX_SAMPLE_SIZE = 0.01
    """Maximum sample size in each dimension (m)."""
    DEFAULT_SENSOR_DIMENSION = 5e-4
    """Default sensor dimension (m)."""
    MIN_SENSOR_DIMENSION = 1e-4
    """Minimum sensor dimension (m)."""
    MAX_SENSOR_DIMENSION = 1e-3
    """Maximum sensor dimension (m)."""
    MIN_XY_SIZE_CUSHION = 5e-4
    """Minimum cushion between sensor dimension and sample size in the X and Y dimensions (m)."""
    MIN_Z_SIZE_CUSHION = 1e-3
    """Minimum cushion between sensor dimension and sample size in the Z dimension (m)."""
    DEFAULT_USE_PROVIDED_THERMAL_PARAMETERS = False
    """Default flag value indicating whether to use user-provided thermal parameters."""
    DEFAULT_COOLING_RATE = 1e6
    """Default cooling rate (K/s)."""
    MIN_COOLING_RATE = 1e5
    """Minimum cooling rate (K/s)."""
    MAX_COOLING_RATE = 1e7
    """Maximum cooling rate (K/s)."""
    DEFAULT_THERMAL_GRADIENT = 1e7
    """Default thermal gradient (K/m)."""
    MIN_THERMAL_GRADIENT = 1e5
    """Minimum thermal gradient (K/m)."""
    MAX_THERMAL_GRADIENT = 1e8
    """Maximum thermal gradient (K/m)."""
    DEFAULT_MELT_POOL_WIDTH = 1.5e-4
    """Default melt pool width (m)."""
    MIN_MELT_POOL_WIDTH = 7.5e-5
    """Minimum melt pool width (m)."""
    MAX_MELT_POOL_WIDTH = 8e-4
    """Maximum melt pool width (m)."""
    DEFAULT_MELT_POOL_DEPTH = 1e-4
    """Default melt pool depth (m)."""
    MIN_MELT_POOL_DEPTH = 1.5e-5
    """Minimum melt pool depth (m)."""
    MAX_MELT_POOL_DEPTH = 8e-4
    """Maximum melt pool depth (m)."""
    DEFAULT_RANDOM_SEED = 0
    """The default random seed, which indicates that a random seed was not provided."""
    MIN_RANDOM_SEED = 1
    """Minimum random seed."""
    MAX_RANDOM_SEED = 2**31 - 1
    """Maximum random seed."""

    def __init__(
        self,
        id: str = "",
        *,
        sample_min_x: float = DEFAULT_POSITION_COORDINATE,
        sample_min_y: float = DEFAULT_POSITION_COORDINATE,
        sample_min_z: float = DEFAULT_POSITION_COORDINATE,
        sample_size_x: float = DEFAULT_SAMPLE_SIZE,
        sample_size_y: float = DEFAULT_SAMPLE_SIZE,
        sample_size_z: float = DEFAULT_SAMPLE_SIZE,
        sensor_dimension: float = DEFAULT_SENSOR_DIMENSION,
        use_provided_thermal_parameters: bool = DEFAULT_USE_PROVIDED_THERMAL_PARAMETERS,
        cooling_rate: float = DEFAULT_COOLING_RATE,
        thermal_gradient: float = DEFAULT_THERMAL_GRADIENT,
        melt_pool_width: float = DEFAULT_MELT_POOL_WIDTH,
        melt_pool_depth: float = DEFAULT_MELT_POOL_DEPTH,
        random_seed: int = DEFAULT_RANDOM_SEED,
        machine: AdditiveMachine = AdditiveMachine(),
        material: AdditiveMaterial = AdditiveMaterial(),
    ):
        """Initialize a ``MicrostructureInput`` object."""

        # we have a circular dependency here, so we validate sensor_dimension
        # and sample_size_* then assign them without calling the setters
        self.__validate_range(
            sensor_dimension,
            self.MIN_SENSOR_DIMENSION,
            self.MAX_SENSOR_DIMENSION,
            "sensor_dimension",
        )
        self.__validate_size(
            sample_size_x, sensor_dimension, self.MIN_XY_SIZE_CUSHION, "sample_size_x"
        )
        self.__validate_size(
            sample_size_y, sensor_dimension, self.MIN_XY_SIZE_CUSHION, "sample_size_y"
        )
        self.__validate_size(
            sample_size_z, sensor_dimension, self.MIN_Z_SIZE_CUSHION, "sample_size_z"
        )
        self._sensor_dimension = sensor_dimension
        self._sample_size_x = sample_size_x
        self._sample_size_y = sample_size_y
        self._sample_size_z = sample_size_z

        # use setters for remaining properties
        self.id = id
        self.sample_min_x = sample_min_x
        self.sample_min_y = sample_min_y
        self.sample_min_z = sample_min_z
        self.use_provided_thermal_parameters = use_provided_thermal_parameters
        self.cooling_rate = cooling_rate
        self.thermal_gradient = thermal_gradient
        self.melt_pool_width = melt_pool_width
        self.melt_pool_depth = melt_pool_depth
        self.machine = machine
        self.material = material
        if random_seed != self.DEFAULT_RANDOM_SEED:
            self.random_seed = random_seed
        else:
            self._random_seed = random_seed

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            if k == "_machine" or k == "_material":
                repr += "\n" + k.replace("_", "", 1) + ": " + str(getattr(self, k))
            else:
                repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, MicrostructureInput):
            return False
        return (
            self.id == __o.id
            and self.sample_min_x == __o.sample_min_x
            and self.sample_min_y == __o.sample_min_y
            and self.sample_min_z == __o.sample_min_z
            and self.sample_size_x == __o.sample_size_x
            and self.sample_size_y == __o.sample_size_y
            and self.sample_size_z == __o.sample_size_z
            and self.sensor_dimension == __o.sensor_dimension
            and self.use_provided_thermal_parameters == __o.use_provided_thermal_parameters
            and self.cooling_rate == __o.cooling_rate
            and self.thermal_gradient == __o.thermal_gradient
            and self.melt_pool_width == __o.melt_pool_width
            and self.melt_pool_depth == __o.melt_pool_depth
            and self.random_seed == __o.random_seed
            and self.machine == __o.machine
            and self.material == __o.material
        )

    @staticmethod
    def __validate_range(value, min, max, name):
        if math.isnan(value):
            raise ValueError("{} must be a number.".format(name))
        if value < min or value > max:
            raise ValueError("{} must be between {} and {}.".format(name, min, max))

    @staticmethod
    def __validate_size(size_value, sensor_value, cushion, name):
        if math.isnan(size_value):
            raise ValueError("{} must be a number.".format(name))
        if size_value - sensor_value < cushion:
            raise ValueError(
                "{} must be at least {} larger than sensor_dimension.".format(name, cushion)
            )

    @property
    def id(self) -> str:
        """User-provided ID for this simulation."""
        return self._id

    @id.setter
    def id(self, value: str):
        self._id = value

    @property
    def machine(self):
        """Machine-related parameters."""
        return self._machine

    @machine.setter
    def machine(self, value):
        self._machine = value

    @property
    def material(self):
        """Material used during simulation."""
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def sample_min_x(self) -> float:
        """Minimum x coordinate of the geometry sample (m).

        Valid values are from the :obj:`MIN_POSITION_COORDINATE` value to
        the :obj:`MAX_POSITION_COORDINATE` value.
        """
        return self._sample_min_x

    @sample_min_x.setter
    def sample_min_x(self, value: float):
        self.__validate_range(
            value, self.MIN_POSITION_COORDINATE, self.MAX_POSITION_COORDINATE, "sample_min_x"
        )
        self._sample_min_x = value

    @property
    def sample_min_y(self) -> float:
        """Minimum y coordinate of the geometry sample (m).

        Valid values are from the :obj:`MIN_POSITION_COORDINATE` value to the
        :obj:`MAX_POSITION_COORDINATE` value.
        """
        return self._sample_min_y

    @sample_min_y.setter
    def sample_min_y(self, value: float):
        self.__validate_range(
            value, self.MIN_POSITION_COORDINATE, self.MAX_POSITION_COORDINATE, "sample_min_y"
        )
        self._sample_min_y = value

    @property
    def sample_min_z(self) -> float:
        """Minimum z coordinate of the geometry sample (m).

        Valid values are from the :obj:`MIN_POSITION_COORDINATE` value to the
        :obj:`MAX_POSITION_COORDINATE` value.
        """
        return self._sample_min_z

    @sample_min_z.setter
    def sample_min_z(self, value: float):
        self.__validate_range(
            value, self.MIN_POSITION_COORDINATE, self.MAX_POSITION_COORDINATE, "sample_min_z"
        )
        self._sample_min_z = value

    @property
    def sample_size_x(self) -> float:
        """Size of the geometry sample in the x direction (m).

        Valid values are from the :obj:`MIN_SAMPLE_SIZE` value to the
        :obj:`MAX_SAMPLE_SIZE` value.
        When setting the```sample_size_x`` parameter, the value must be greater than the
        ``sensor_dimension`` value plus the :obj:`MIN_XY_SIZE_CUSHION` value.
        """
        return self._sample_size_x

    @sample_size_x.setter
    def sample_size_x(self, value: float):
        self.__validate_range(value, self.MIN_SAMPLE_SIZE, self.MAX_SAMPLE_SIZE, "sample_size_x")
        self.__validate_size(
            value, self.sensor_dimension, self.MIN_XY_SIZE_CUSHION, "sample_size_x"
        )
        self._sample_size_x = value

    @property
    def sample_size_y(self) -> float:
        """Size of the geometry sample in the y direction (m).

        Valid values are from the :obj:`MIN_SAMPLE_SIZE` value to the
        :obj:`MAX_SAMPLE_SIZE` value.
        When setting the ``sample_size_y`` parameter, the value must be greater than the
        ``sensor_dimension`` value plus the :obj:`MIN_XY_SIZE_CUSHION` value.
        """
        return self._sample_size_y

    @sample_size_y.setter
    def sample_size_y(self, value: float):
        self.__validate_range(value, self.MIN_SAMPLE_SIZE, self.MAX_SAMPLE_SIZE, "sample_size_y")
        self.__validate_size(
            value, self.sensor_dimension, self.MIN_XY_SIZE_CUSHION, "sample_size_y"
        )
        self._sample_size_y = value

    @property
    def sample_size_z(self) -> float:
        """Size of the geometry sample in the z direction (m).

        Valid values are from the :obj:`MIN_SAMPLE_SIZE` value to the
        :obj:`MAX_SAMPLE_SIZE` value.
        When setting the ``sample_size_y`` parameter, the value must be greater than the
        ``sensor_dimension`` value plus the :obj:`MIN_XY_SIZE_CUSHION` value.
        """
        return self._sample_size_z

    @sample_size_z.setter
    def sample_size_z(self, value: float):
        self.__validate_range(value, self.MIN_SAMPLE_SIZE, self.MAX_SAMPLE_SIZE, "sample_size_z")
        self.__validate_size(value, self.sensor_dimension, self.MIN_Z_SIZE_CUSHION, "sample_size_z")
        self._sample_size_z = value

    @property
    def sensor_dimension(self) -> float:
        """Dimension of the sensor (m).

        Valid values are from the :obj:`MIN_SENSOR_DIMENSION` value to the
        :obj:`MAX_SENSOR_DIMENSION` value.
        """
        return self._sensor_dimension

    @sensor_dimension.setter
    def sensor_dimension(self, value: float):
        self.__validate_range(
            value, self.MIN_SENSOR_DIMENSION, self.MAX_SENSOR_DIMENSION, "sensor_dimension"
        )
        size_errors = ""
        try:
            self.__validate_size(
                self.sample_size_x, value, self.MIN_XY_SIZE_CUSHION, "sample_size_x"
            )
        except ValueError as e:
            size_errors += str(e) + "\n"
        try:
            self.__validate_size(
                self.sample_size_y, value, self.MIN_XY_SIZE_CUSHION, "sample_size_y"
            )
        except ValueError as e:
            size_errors += str(e) + "\n"
        try:
            self.__validate_size(
                self.sample_size_z, value, self.MIN_Z_SIZE_CUSHION, "sample_size_z"
            )
        except ValueError as e:
            size_errors += str(e) + "\n"
        if size_errors:
            raise ValueError(size_errors)

        self._sensor_dimension = value

    @property
    def use_provided_thermal_parameters(self) -> bool:
        """Flag indicating if the ``cooling_rate``, ``thermal_gradient``, ``melt_pool_depth``, and ``melt_pool_width`` parameters have been provided by the user.

        If the value is ``False``, these values
        will be calculated. Default is ``False``.
        """  # noqa: E501
        return self._use_provided_thermal_parameters

    @use_provided_thermal_parameters.setter
    def use_provided_thermal_parameters(self, value: bool):
        self._use_provided_thermal_parameters = value

    @property
    def cooling_rate(self) -> float:
        """Material cooling rate (K/s).

        Valid values are from the :obj:`MIN_COOLING_RATE` value to the
        :obj:`MAX_COOLING_RATE` value.
        """
        return self._cooling_rate

    @cooling_rate.setter
    def cooling_rate(self, value: float):
        self.__validate_range(value, self.MIN_COOLING_RATE, self.MAX_COOLING_RATE, "cooling_rate")
        self._cooling_rate = value

    @property
    def thermal_gradient(self) -> float:
        """Material thermal gradient (K/m).

        Valid values are from the :obj:`MIN_THERMAL_GRADIENT` value to the
        :obj:`MAX_THERMAL_GRADIENT` value.
        """
        return self._thermal_gradient

    @thermal_gradient.setter
    def thermal_gradient(self, value: float):
        self.__validate_range(
            value, self.MIN_THERMAL_GRADIENT, self.MAX_THERMAL_GRADIENT, "thermal_gradient"
        )
        self._thermal_gradient = value

    @property
    def melt_pool_width(self) -> float:
        """Melt pool width (m).

        This is the width of the melt pool measured at the top of the powder layer.

        Valid values are from the :obj:`MIN_MELT_POOL_WIDTH` value to the
        :obj:`MAX_MELT_POOL_WIDTH` value.
        """
        return self._melt_pool_width

    @melt_pool_width.setter
    def melt_pool_width(self, value: float):
        self.__validate_range(
            value, self.MIN_MELT_POOL_WIDTH, self.MAX_MELT_POOL_WIDTH, "melt_pool_width"
        )
        self._melt_pool_width = value

    @property
    def melt_pool_depth(self) -> float:
        """Melt pool depth (m).

        This is the depth of the melt pool as measured from the top of the powder layer.

        Valid values are from the :obj:`MIN_MELT_POOL_DEPTH` value to the
        :obj:`MAX_MELT_POOL_DEPTH` value.
        """
        return self._melt_pool_depth

    @melt_pool_depth.setter
    def melt_pool_depth(self, value: float):
        self.__validate_range(
            value, self.MIN_MELT_POOL_DEPTH, self.MAX_MELT_POOL_DEPTH, "melt_pool_depth"
        )
        self._melt_pool_depth = value

    @property
    def random_seed(self) -> int:
        """Random seed for the simulation.

        Valid values are from the :obj:`MIN_RANDOM_SEED` value to the
        :obj:`MAX_RANDOM_SEED` value.
        """
        return self._random_seed

    @random_seed.setter
    def random_seed(self, value: int):
        self.__validate_range(value, self.MIN_RANDOM_SEED, self.MAX_RANDOM_SEED, "random_seed")
        self._random_seed = value

    def _to_simulation_request(self) -> SimulationRequest:
        """Convert this object into a simulation request message."""
        input = MicrostructureInputMessage(
            machine=self.machine._to_machine_message(),
            material=self.material._to_material_message(),
            cube_min_x=self.sample_min_x,
            cube_min_y=self.sample_min_y,
            cube_min_z=self.sample_min_z,
            cube_size_x=self.sample_size_x,
            cube_size_y=self.sample_size_y,
            cube_size_z=self.sample_size_z,
            sensor_dimension=self.sensor_dimension,
            use_provided_thermal_parameters=self.use_provided_thermal_parameters,
            cooling_rate=self.cooling_rate,
            thermal_gradient=self.thermal_gradient,
            melt_pool_width=self.melt_pool_width,
            melt_pool_depth=self.melt_pool_depth,
            use_random_seed=(self.random_seed != self.DEFAULT_RANDOM_SEED),
            random_seed=self.random_seed,
        )
        return SimulationRequest(id=self.id, microstructure_input=input)


class CircleEquivalenceColumnNames:
    """Provides column names for the circle equivalence data frame."""

    GRAIN_NUMBER = "grain_number"
    """Grain number."""
    AREA_FRACTION = "area_fraction"
    """Area fraction for grain."""
    DIAMETER = "diameter_um"
    """Grain diameter (µm)."""
    ORIENTATION_ANGLE = "orientation_angle"
    """Orientation angle (degrees)."""


class MicrostructureSummary:
    """Provides the summary of a microstructure simulation."""

    def __init__(
        self, input: MicrostructureInput, result: MicrostructureResultMessage, user_data_path: str
    ) -> None:
        """Initialize a ``MicrostructureSummary`` object."""
        if not isinstance(input, MicrostructureInput):
            raise ValueError("Invalid input type passed to init, " + self.__class__.__name__)
        if not isinstance(result, MicrostructureResultMessage):
            raise ValueError("Invalid result type passed to init, " + self.__class__.__name__)
        if not user_data_path:
            raise ValueError("Invalid user data path, " + self.__class__.__name__)
        self._input = input
        id = input.id if input.id else misc.short_uuid()
        outpath = os.path.join(user_data_path, id)
        self._result = _Microstructure2DResult(result, outpath)

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in [x for x in self.__dict__ if x != "_result"]:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        repr += self._result.__repr__()
        return repr

    @property
    def input(self):
        """Simulation input.

        For more information, see the :class:`MicrostructureInput` class.
        """
        return self._input

    @property
    def xy_vtk(self) -> str:
        """Path to the VTK file containing the 2D grain structure data in the XY plane.

        The VTK file contains these scalar data sets: ``GrainBoundaries``, ``Orientation_(deg)``, and
        ``GrainNumber``.
        """
        return self._result._xy_vtk

    @property
    def xz_vtk(self) -> str:
        """Path to the VTK file containing the 2D grain structure data in the XZ plane.

        The VTK file contains these scalar data sets: ``GrainBoundaries``,
        ``Orientation_(deg)``, and ``GrainNumber``.
        """
        return self._result._xz_vtk

    @property
    def yz_vtk(self) -> str:
        """Path to the VTK file containing the 2D grain structure data in the YZ plane.

        The VTK file contains these scalar data sets: ``GrainBoundaries``,
        ``Orientation_(deg)``, and ``GrainNumber``.
        """
        return self._result._yz_vtk

    @property
    def xy_circle_equivalence(self) -> pd.DataFrame:
        """Circle equivalence data for the XY plane.

        For data frame column names, see the :class:`CircleEquivalenceColumnNames` class.
        """
        return self._result._xy_circle_equivalence

    @property
    def xz_circle_equivalence(self) -> pd.DataFrame:
        """Circle equivalence data for the XZ plane.

        For data frame column names, see the :class:`CircleEquivalenceColumnNames` class.
        """
        return self._result._xz_circle_equivalence

    @property
    def yz_circle_equivalence(self) -> pd.DataFrame:
        """Circle equivalence data for the YZ plane.

        For data frame column names, see the :class:`CircleEquivalenceColumnNames` class.
        """
        return self._result._yz_circle_equivalence

    @property
    def xy_average_grain_size(self) -> float:
        """Average grain size (µm) for the XY plane."""
        return self._result._xy_average_grain_size

    @property
    def xz_average_grain_size(self) -> float:
        """Average grain size (µm) for the XZ plane."""
        return self._result._xz_average_grain_size

    @property
    def yz_average_grain_size(self) -> float:
        """Average grain size (µm) for the YZ plane."""
        return self._result._yz_average_grain_size

    @staticmethod
    def _circle_equivalence_frame(src: RepeatedCompositeFieldContainer) -> pd.DataFrame:
        d = {}
        d[CircleEquivalenceColumnNames.GRAIN_NUMBER] = np.asarray([x.grain_number for x in src])
        d[CircleEquivalenceColumnNames.AREA_FRACTION] = np.asarray([x.area_fraction for x in src])
        d[CircleEquivalenceColumnNames.DIAMETER] = np.asarray([x.diameter_um for x in src])
        d[CircleEquivalenceColumnNames.ORIENTATION_ANGLE] = np.asarray(
            [math.degrees(x.orientation_angle) for x in src]
        )
        return pd.DataFrame(d)

    @staticmethod
    def _average_grain_size(df: pd.DataFrame) -> float:
        """Calculate the average grain size (µm) for a given plane.

        Parameters
        ----------
        df : pd.DataFrame
            Data frame containing circle equivalence data for a given plane.

        Returns
        -------
        float
            Average grain size (µm).
        """

        return (
            df[CircleEquivalenceColumnNames.DIAMETER]
            * df[CircleEquivalenceColumnNames.AREA_FRACTION]
        ).sum()


class _Microstructure2DResult:
    """Provides the results of a 2D microstructure simulation."""

    def __init__(self, msg: MicrostructureResultMessage, output_data_path: str) -> None:
        """Initialize a ``Microstructure2DResult`` object."""
        if not isinstance(msg, MicrostructureResultMessage):
            raise ValueError("Invalid msg parameter, " + self.__class__.__name__)
        if not output_data_path:
            raise ValueError("Invalid output data path, " + self.__class__.__name__)

        if not os.path.exists(output_data_path):
            os.makedirs(output_data_path)
        self._xy_vtk = os.path.join(output_data_path, "xy.vtk")
        with open(self._xy_vtk, "wb") as xy_vtk:
            xy_vtk.write(msg.xy_vtk)
        self._xz_vtk = os.path.join(output_data_path, "xz.vtk")
        with open(self._xz_vtk, "wb") as xz_vtk:
            xz_vtk.write(msg.xz_vtk)
        self._yz_vtk = os.path.join(output_data_path, "yz.vtk")
        with open(self._yz_vtk, "wb") as yz_vtk:
            yz_vtk.write(msg.yz_vtk)

        self._xy_circle_equivalence = MicrostructureSummary._circle_equivalence_frame(
            msg.xy_circle_equivalence
        )
        self._xz_circle_equivalence = MicrostructureSummary._circle_equivalence_frame(
            msg.xz_circle_equivalence
        )
        self._yz_circle_equivalence = MicrostructureSummary._circle_equivalence_frame(
            msg.yz_circle_equivalence
        )
        self._xy_average_grain_size = MicrostructureSummary._average_grain_size(
            self._xy_circle_equivalence
        )
        self._xz_average_grain_size = MicrostructureSummary._average_grain_size(
            self._xz_circle_equivalence
        )
        self._yz_average_grain_size = MicrostructureSummary._average_grain_size(
            self._yz_circle_equivalence
        )

    def __repr__(self):
        repr = ""
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides logging and progress."""

import logging
from os import getenv

from ansys.api.additive.v0.additive_domain_pb2 import Progress, ProgressState
from tqdm import tqdm


class ProgressLogger:
    """Provides for logging and progress reporting."""

    def __init__(self, name: str = None) -> None:
        """Initialize a ``ProgressLogger`` object.

        Parameters
        ----------
        name: str
            Name of the logger.
        """
        self._log = logging.getLogger(name)
        self._last_percent_complete = 0
        self._last_context = "Initializing"

    def log_progress(self, progress: Progress):
        """Report progress and possibly emit a log message.

        Parameters
        ----------
        progress: Progress
            Latest progress.
        """
        if getenv("GENERATING_DOCS"):
            # Don't send  progress when generating docs
            return
        if not hasattr(self, "_pbar"):
            self._pbar = tqdm(
                total=100,
                colour="green",
                desc=self._last_context,
                dynamic_ncols=True,
            )

        if progress.message and "SOLVERINFO" in progress.message:
            self._log.debug(progress.message)
            return

        if progress.state == ProgressState.PROGRESS_STATE_ERROR:
            self._pbar.write(progress.message)
            return

        if progress.context and progress.context != self._last_context:
            if "Solving Layer" not in progress.context or progress.context == "Solving Layer 1":
                self._pbar.reset(total=100)
                self._pbar.set_description(progress.context)
                self._last_context = progress.context
                self._last_percent_complete = 0
            else:
                self._pbar.set_description(progress.context, refresh=False)

        if progress.percent_complete - self._last_percent_complete > 0:
            self._pbar.update(progress.percent_complete - self._last_percent_complete)
        self._last_percent_complete = progress.percent_complete

    def __del__(self):
        if hasattr(self, "_pbar"):
            self._pbar.close()



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Python client for the Ansys Additive service."""
import os

import platformdirs

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata  # type: ignore

__version__ = importlib_metadata.version(__name__.replace(".", "-"))

__APP_NAME = "pyadditive"
__COMPANY_NAME = "Ansys Inc"

# Setup data directory
USER_DATA_PATH = platformdirs.user_data_dir(__APP_NAME, __COMPANY_NAME)
"""Storage path for user data."""
if not os.path.exists(USER_DATA_PATH):  # pragma: no cover
    os.makedirs(USER_DATA_PATH)

EXAMPLES_PATH = os.path.join(USER_DATA_PATH, "examples")
"""Storage path for example data."""
if not os.path.exists(EXAMPLES_PATH):  # pragma: no cover
    os.makedirs(EXAMPLES_PATH)

from ansys.additive.core.additive import Additive
from ansys.additive.core.exceptions import BetaFeatureNotEnabledError
from ansys.additive.core.geometry_file import BuildFile, MachineType, StlFile
from ansys.additive.core.logger import LOG
from ansys.additive.core.machine import AdditiveMachine, MachineConstants
from ansys.additive.core.material import (
    AdditiveMaterial,
    CharacteristicWidthDataPoint,
    ThermalPropertiesDataPoint,
)
from ansys.additive.core.material_tuning import MaterialTuningInput, MaterialTuningSummary
from ansys.additive.core.microstructure import (
    CircleEquivalenceColumnNames,
    MicrostructureInput,
    MicrostructureSummary,
)
from ansys.additive.core.microstructure_3d import Microstructure3DInput, Microstructure3DSummary
from ansys.additive.core.porosity import PorosityInput, PorositySummary
from ansys.additive.core.simulation import SimulationError, SimulationStatus, SimulationType
from ansys.additive.core.single_bead import (
    MeltPool,
    MeltPoolColumnNames,
    SingleBeadInput,
    SingleBeadSummary,
)
from ansys.additive.core.thermal_history import (
    CoaxialAverageSensorInputs,
    Range,
    ThermalHistoryInput,
    ThermalHistorySummary,
)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides input and result summary containers for porosity simulations."""
import math

from ansys.api.additive.v0.additive_domain_pb2 import PorosityInput as PorosityInputMessage
from ansys.api.additive.v0.additive_domain_pb2 import PorosityResult
from ansys.api.additive.v0.additive_simulation_pb2 import SimulationRequest

from ansys.additive.core.machine import AdditiveMachine
from ansys.additive.core.material import AdditiveMaterial


class PorosityInput:
    """Provides input parameters for porosity simulation."""

    DEFAULT_SAMPLE_SIZE = 3e-3
    """Default sample size in each dimension (m)."""
    MIN_SAMPLE_SIZE = 1e-3
    """Minimum sample size in each dimension (m)."""
    MAX_SAMPLE_SIZE = 1e-2
    """Maximum sample size in each dimension (m)."""

    def __init__(
        self,
        id: str = "",
        *,
        size_x: float = DEFAULT_SAMPLE_SIZE,
        size_y: float = DEFAULT_SAMPLE_SIZE,
        size_z: float = DEFAULT_SAMPLE_SIZE,
        machine: AdditiveMachine = AdditiveMachine(),
        material: AdditiveMaterial = AdditiveMaterial(),
    ):
        """Initialize a ``PorosityInput`` object."""
        self.id = id
        self.size_x = size_x
        self.size_y = size_y
        self.size_z = size_z
        self.machine = machine
        self.material = material

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            if k == "_machine" or k == "_material":
                repr += "\n" + k.replace("_", "", 1) + ": " + str(getattr(self, k))
            else:
                repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, PorosityInput):
            return False
        return (
            self.id == __o.id
            and self.size_x == __o.size_x
            and self.size_y == __o.size_y
            and self.size_z == __o.size_z
            and self.machine == __o.machine
            and self.material == __o.material
        )

    def __validate_range(self, value, min, max, name):
        if math.isnan(value):
            raise ValueError("{} must be a number.".format(name))
        if value < min or value > max:
            raise ValueError("{} must be between {} and {}.".format(name, min, max))

    @property
    def id(self) -> str:
        """User-provided ID for the simulation."""
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def machine(self) -> AdditiveMachine:
        """Machine-related parameters."""
        return self._machine

    @machine.setter
    def machine(self, value):
        self._machine = value

    @property
    def material(self) -> AdditiveMaterial:
        """Material used during the simulation."""
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def size_x(self) -> float:
        """Size (m) of the simulated sample in the x dimension.

        Valid values are from the :obj:`MIN_SAMPLE_SIZE` value to the
        :obj:`MAX_SAMPLE_SIZE` value.
        """
        return self._size_x

    @size_x.setter
    def size_x(self, value):
        self.__validate_range(value, self.MIN_SAMPLE_SIZE, self.MAX_SAMPLE_SIZE, "size_x")
        self._size_x = value

    @property
    def size_y(self) -> float:
        """Size (m) of the simulated sample in the y dimension.

        Valid values are from the :obj:`MIN_SAMPLE_SIZE` value to the
        :obj:`MAX_SAMPLE_SIZE` value.
        """
        return self._size_y

    @size_y.setter
    def size_y(self, value):
        self.__validate_range(value, self.MIN_SAMPLE_SIZE, self.MAX_SAMPLE_SIZE, "size_y")
        self._size_y = value

    @property
    def size_z(self) -> float:
        """Size (m) of the simulated sample in the z dimension.

        Valid values are from the :obj:`MIN_SAMPLE_SIZE` value to the
        :obj:`MAX_SAMPLE_SIZE` value.
        """
        return self._size_z

    @size_z.setter
    def size_z(self, value):
        self.__validate_range(value, self.MIN_SAMPLE_SIZE, self.MAX_SAMPLE_SIZE, "size_z")
        self._size_z = value

    def _to_simulation_request(self) -> SimulationRequest:
        """Convert this object into a simulation request message."""
        input = PorosityInputMessage(
            machine=self.machine._to_machine_message(),
            material=self.material._to_material_message(),
            size_x=self.size_x,
            size_y=self.size_y,
            size_z=self.size_z,
        )
        return SimulationRequest(id=self.id, porosity_input=input)


class PorositySummary:
    """Provides a summary of a porosity simulation."""

    def __init__(
        self,
        input: PorosityInput,
        result: PorosityResult,
    ):
        """Initialize a ``PorositySummary`` object."""
        if not isinstance(input, PorosityInput):
            raise ValueError("Invalid input type passed to init, " + self.__class__.__name__)
        if not isinstance(result, PorosityResult):
            raise ValueError("Invalid result type passed to init, " + self.__class__.__name__)
        self._input = input
        self._relative_density = result.solid_ratio

    @property
    def input(self) -> PorosityInput:
        """Simulation input."""
        return self._input

    @property
    def relative_density(self) -> float:
        """Ratio of the density of the simulated sample to a completely solid sample."""
        return self._relative_density

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides common definitions and classes for simulations."""
from typing import Union

from ansys.additive.core.microstructure import MicrostructureInput
from ansys.additive.core.porosity import PorosityInput
from ansys.additive.core.single_bead import SingleBeadInput
from ansys.additive.core.thermal_history import ThermalHistoryInput


class SimulationType:
    """Provides simulation types."""

    SINGLE_BEAD = "SingleBead"
    """Single bead simulation."""
    POROSITY = "Porosity"
    """Porosity simulation."""
    MICROSTRUCTURE = "Microstructure"
    """Microstructure simulation."""


class SimulationStatus:
    """Provides simulation status values."""

    PENDING = "Pending"
    """Simulation is waiting to run."""
    COMPLETED = "Completed"
    """Simulation was run."""
    ERROR = "Error"
    """Simulation errored."""
    SKIP = "Skip"
    """Do not run this simulation, only applies to parametric studies."""


class SimulationError:
    """Provides simulation errors."""

    def __init__(
        self,
        input: Union[SingleBeadInput, PorosityInput, MicrostructureInput, ThermalHistoryInput],
        message: str,
    ):
        """Initialize a ``SimulationError`` object."""
        self._input = input
        self._message = message

    @property
    def input(
        self,
    ) -> Union[SingleBeadInput, PorosityInput, MicrostructureInput, ThermalHistoryInput]:
        """Simulation input."""
        return self._input

    @property
    def message(self) -> str:
        """Provides simulation error message."""
        return self._message



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a container for part geometry files."""
from enum import IntEnum
import os

from ansys.api.additive.v0.additive_domain_pb2 import BuildFileMachineType


class MachineType(IntEnum):
    """Machine type values."""

    NONE = BuildFileMachineType.BUILD_FILE_MACHINE_TYPE_NONE
    ADDITIVE_INDUSTRIES = BuildFileMachineType.BUILD_FILE_MACHINE_TYPE_AI
    SLM = BuildFileMachineType.BUILD_FILE_MACHINE_TYPE_SLM
    RENISHAW = BuildFileMachineType.BUILD_FILE_MACHINE_TYPE_RENISHAW
    EOS = BuildFileMachineType.BUILD_FILE_MACHINE_TYPE_EOS
    TRUMPF = BuildFileMachineType.BUILD_FILE_MACHINE_TYPE_TRUMPF
    HB3D = BuildFileMachineType.BUILD_FILE_MACHINE_TYPE_HB3D
    SISMA = BuildFileMachineType.BUILD_FILE_MACHINE_TYPE_SISMA


class BuildFile:
    """Provides the build file description.

    In the context of PyAdditive, a build file is a ZIP archive
    containing:

    - A part geometry file in STL format.
    - Zero or more support geometry files in STL format. Support file
      names must end in ``*_vless.stl`` for volumeless supports and
      ``*_solid.stl`` for solid supports.
    - One or more machine instruction files. The number and type of
      files depend upon the machine type.

    These files must be placed in the root of the archive and not
    under a folder.

    Parameters
    ----------
    type: MachineType
        Type of additive machine the build file is for.
    path: os.PathLike
        Path to the build file.
    """

    def __init__(self, type: MachineType, path: os.PathLike):
        """Initialize a ``BuildFile`` object."""
        if not isinstance(type, MachineType):
            raise ValueError("Invalid machine type")
        if not os.path.exists(path):
            raise FileNotFoundError(f"File does not exist, {path}")
        self._type = type
        self._path = path

    def __repr__(self):
        repr = (
            type(self).__name__
            + "\n"
            + f"type: {MachineType(self._type).name}\n"
            + f"path: {self._path}\n"
        )
        return repr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BuildFile):
            return False
        for k in self.__dict__:
            if getattr(self, k) != getattr(other, k):
                return False
        return True

    @property
    def type(self) -> MachineType:
        """Additive manufacturing machine type this file is for."""
        return self._type

    @type.setter
    def type(self, value: MachineType):
        """Set machine type."""
        if not isinstance(value, MachineType):
            raise ValueError("Attempted to assign type with invalid value")
        self._type = value

    @property
    def path(self) -> str:
        """Path of the ZIP file."""
        return self._path

    @path.setter
    def path(self, value: os.PathLike):
        """Set file path."""
        if not os.path.exists(value):
            raise FileNotFoundError(f"File does not exist, {value}")
        self._path = value


class StlFile:
    """Container for the STL file definition."""

    def __init__(self, path: str):
        """Initialize an ``StlFile`` object.

        Parameters
        ----------
        path: str
            Path to file.
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"File does not exist, {path}")
        self._path = path

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StlFile):
            return False
        for k in self.__dict__:
            if getattr(self, k) != getattr(other, k):
                return False
        return True

    @property
    def path(self) -> str:
        """Path of the STL file."""
        return self._path

    @path.setter
    def path(self, value: os.PathLike):
        """Set file path."""
        if not os.path.exists(value):
            raise FileNotFoundError(f"File does not exist, {value}")
        self._path = value



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides startup utilities for a local Additive server."""
from __future__ import annotations

from datetime import datetime
import os
from pathlib import Path
import socket
import subprocess
import time

from ansys.additive.core import USER_DATA_PATH
from ansys.additive.core.server_connection.constants import (
    ADDITIVE_SERVER_EXE_NAME,
    ADDITIVE_SERVER_SUBDIR,
    DEFAULT_PRODUCT_VERSION,
)


class LocalServer:
    """Provides startup utilities for a local Additive server."""

    @staticmethod
    def launch(
        port: int, cwd: str = USER_DATA_PATH, product_version: str = DEFAULT_PRODUCT_VERSION
    ) -> subprocess.Popen:
        """Launch a local gRPC server for the Additive service.

        Parameters
        ----------
        port: int
            Port number to use for gRPC connections.
        cwd: str
            Current working directory to use for the server process.
        product_version: str
            Version of the Ansys installation to use, of the form ``"YYR"``, where
            ``YY`` is the two digit year and ``R`` is the release for
            that year. For example, Ansys 2024 R1 would be ``"241"``.

        Returns
        -------
        process: subprocess.Popen
            Server process. To stop the server, call the ``kill()`` method on the returned object.
        """
        server_exe = ""
        if os.name == "nt":
            env_var = (
                f"AWP_ROOT{product_version}"
                if product_version
                else f"AWP_ROOT{DEFAULT_PRODUCT_VERSION}"
            )
            awp_root = os.environ.get(env_var)
            if not awp_root:
                raise FileNotFoundError("Cannot find Ansys installation directory")
            server_exe = Path(awp_root) / ADDITIVE_SERVER_SUBDIR / f"{ADDITIVE_SERVER_EXE_NAME}.exe"
        elif os.name == "posix":
            base_path = None
            for path in ["/usr/ansys_inc", "/ansys_inc"]:
                if os.path.isdir(path):
                    base_path = path
                    break
            if not base_path:
                raise FileNotFoundError("Cannot find Ansys installation directory")
            server_exe = (
                Path(base_path)
                / f"v{product_version}"
                / ADDITIVE_SERVER_SUBDIR
                / ADDITIVE_SERVER_EXE_NAME
            )
        else:
            raise OSError(f"Unsupported OS {os.name}")

        if not server_exe.exists():
            raise FileNotFoundError(f"Cannot find {server_exe}")

        Path(cwd).mkdir(parents=True, exist_ok=True)

        start_time = datetime.now().strftime("%Y%m%d_%H%M%S")

        with open(os.path.join(cwd, f"additiveserver_{start_time}.log"), "w") as log_file:
            server_process = subprocess.Popen(
                f'"{server_exe}" --port {port}',
                shell=os.name != "nt",  # use shell on Linux
                cwd=cwd,
                stdout=log_file,
                stderr=subprocess.STDOUT,
            )
            time.sleep(2)  # wait for server to start
            if server_process.poll():
                raise Exception(f"Server exited with code {server_process.returncode}")

        return server_process

    @staticmethod
    def find_open_port() -> int:
        """Find an open port on the local host.

        Returns
        -------
        port: int
            Open port number.

        .. note::
            This port may be taken by the time you try to use it.
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        port = s.getsockname()[1]
        s.close()
        return port



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a class to run parametric study simulations."""
from __future__ import annotations

import numpy as np
import pandas as pd

from ansys.additive.core import (
    Additive,
    AdditiveMachine,
    AdditiveMaterial,
    MachineConstants,
    MicrostructureInput,
    MicrostructureSummary,
    PorosityInput,
    PorositySummary,
    SimulationStatus,
    SimulationType,
    SingleBeadInput,
    SingleBeadSummary,
)
from ansys.additive.core.parametric_study.constants import ColumnNames


class ParametricRunner:
    """Provides methods to run parametric study simulations."""

    @staticmethod
    def simulate(
        df: pd.DataFrame,
        additive: Additive,
        type: list[SimulationType] = None,
        priority: int = None,
        iteration: int = None,
    ) -> list[SingleBeadSummary, PorositySummary, MicrostructureSummary]:
        """Run the simulations in the parametric study with ``Status`` equal to ``Pending``.

        Execution order is determined by the ``Priority`` value assigned to the simulations.
        Lower values are interpreted as having higher priority and are run first.

        Parameters
        ----------
        df : pd.DataFrame
            Parametric study data frame.
        additive : Additive
            Additive service connection to use for running simulations.
        type : list, default: None
            List of the simulation types to run. The default is ``None``, in which case all
            simulation types are run.
        priority : int, default: None
            Priority of simulations to run. The default is ``None``, in which case
            all priorities are run.
        iteration : int, default: None
            Iteration number of simulations to run. The default is ``None``, in which case
            all iterations are run.

        Returns
        -------
        list[SingleBeadSummary, PorositySummary, MicrostructureSummary]
            List of simulation summaries.
        """
        if type is None:
            type = [
                SimulationType.SINGLE_BEAD,
                SimulationType.POROSITY,
                SimulationType.MICROSTRUCTURE,
            ]
        elif not isinstance(type, list):
            type = [type]

        view = df[
            (df[ColumnNames.STATUS] == SimulationStatus.PENDING) & df[ColumnNames.TYPE].isin(type)
        ]
        if priority is not None:
            view = view[view[ColumnNames.PRIORITY] == priority]
        view = view.sort_values(by=ColumnNames.PRIORITY, ascending=True)

        if iteration is not None:
            view = df[(df[ColumnNames.ITERATION] == iteration)]

        inputs = []
        # NOTICE: We use iterrows() instead of itertuples() here to
        # access values by column name
        for _, row in view.iterrows():
            try:
                material = additive.material(row[ColumnNames.MATERIAL])
            except Exception:
                print(
                    f"Material {row[ColumnNames.MATERIAL]} not found, skipping {row[ColumnNames.ID]}"
                )
                continue
            machine = ParametricRunner._create_machine(row)
            sim_type = row[ColumnNames.TYPE]
            if sim_type == SimulationType.SINGLE_BEAD:
                inputs.append(ParametricRunner._create_single_bead_input(row, material, machine))
            elif sim_type == SimulationType.POROSITY:
                inputs.append(ParametricRunner._create_porosity_input(row, material, machine))
            elif sim_type == SimulationType.MICROSTRUCTURE:
                inputs.append(ParametricRunner._create_microstructure_input(row, material, machine))
            else:  # pragma: no cover
                print(
                    f"Invalid simulation type: {row[ColumnNames.TYPE]} for {row[ColumnNames.ID]}, skipping"
                )
                continue

        summaries = additive.simulate(inputs)

        # TODO: Return the summaries one at a time, possibly as an iterator,
        # so that the data frame can be updated as each simulation completes.
        return summaries

    @staticmethod
    def _create_machine(row: pd.Series) -> AdditiveMachine:
        return AdditiveMachine(
            laser_power=row[ColumnNames.LASER_POWER],
            scan_speed=row[ColumnNames.SCAN_SPEED],
            layer_thickness=row[ColumnNames.LAYER_THICKNESS],
            beam_diameter=row[ColumnNames.BEAM_DIAMETER],
            heater_temperature=row[ColumnNames.HEATER_TEMPERATURE],
            starting_layer_angle=row[ColumnNames.START_ANGLE]
            if not np.isnan(row[ColumnNames.START_ANGLE])
            else MachineConstants.DEFAULT_STARTING_LAYER_ANGLE,
            layer_rotation_angle=row[ColumnNames.ROTATION_ANGLE]
            if not np.isnan(row[ColumnNames.ROTATION_ANGLE])
            else MachineConstants.DEFAULT_LAYER_ROTATION_ANGLE,
            hatch_spacing=row[ColumnNames.HATCH_SPACING]
            if not np.isnan(row[ColumnNames.HATCH_SPACING])
            else MachineConstants.DEFAULT_HATCH_SPACING,
            slicing_stripe_width=row[ColumnNames.STRIPE_WIDTH]
            if not np.isnan(row[ColumnNames.STRIPE_WIDTH])
            else MachineConstants.DEFAULT_SLICING_STRIPE_WIDTH,
        )

    @staticmethod
    def _create_single_bead_input(
        row: pd.Series, material: AdditiveMaterial, machine: AdditiveMachine
    ) -> SingleBeadInput:
        return SingleBeadInput(
            id=row[ColumnNames.ID],
            material=material,
            machine=machine,
            bead_length=row[ColumnNames.SINGLE_BEAD_LENGTH],
        )

    @staticmethod
    def _create_porosity_input(
        row: pd.Series, material: AdditiveMaterial, machine: AdditiveMachine
    ) -> PorosityInput:
        return PorosityInput(
            id=row[ColumnNames.ID],
            material=material,
            machine=machine,
            size_x=row[ColumnNames.POROSITY_SIZE_X],
            size_y=row[ColumnNames.POROSITY_SIZE_Y],
            size_z=row[ColumnNames.POROSITY_SIZE_Z],
        )

    @staticmethod
    def _create_microstructure_input(
        row: pd.Series, material: AdditiveMaterial, machine: AdditiveMachine
    ) -> MicrostructureInput:
        use_provided_thermal_param = (
            not np.isnan(row[ColumnNames.COOLING_RATE])
            or not np.isnan(row[ColumnNames.THERMAL_GRADIENT])
            or not np.isnan(row[ColumnNames.MICRO_MELT_POOL_WIDTH])
            or not np.isnan(row[ColumnNames.MICRO_MELT_POOL_DEPTH])
        )

        return MicrostructureInput(
            id=row[ColumnNames.ID],
            material=material,
            machine=machine,
            sample_size_x=row[ColumnNames.MICRO_SIZE_X],
            sample_size_y=row[ColumnNames.MICRO_SIZE_Y],
            sample_size_z=row[ColumnNames.MICRO_SIZE_Z],
            sensor_dimension=row[ColumnNames.MICRO_SENSOR_DIM],
            use_provided_thermal_parameters=use_provided_thermal_param,
            sample_min_x=row[ColumnNames.MICRO_MIN_X]
            if not np.isnan(row[ColumnNames.MICRO_MIN_X])
            else MicrostructureInput.DEFAULT_POSITION_COORDINATE,
            sample_min_y=row[ColumnNames.MICRO_MIN_Y]
            if not np.isnan(row[ColumnNames.MICRO_MIN_Y])
            else MicrostructureInput.DEFAULT_POSITION_COORDINATE,
            sample_min_z=row[ColumnNames.MICRO_MIN_Z]
            if not np.isnan(row[ColumnNames.MICRO_MIN_Z])
            else MicrostructureInput.DEFAULT_POSITION_COORDINATE,
            cooling_rate=row[ColumnNames.COOLING_RATE]
            if not np.isnan(row[ColumnNames.COOLING_RATE])
            else MicrostructureInput.DEFAULT_COOLING_RATE,
            thermal_gradient=row[ColumnNames.THERMAL_GRADIENT]
            if not np.isnan(row[ColumnNames.THERMAL_GRADIENT])
            else MicrostructureInput.DEFAULT_THERMAL_GRADIENT,
            melt_pool_width=row[ColumnNames.MICRO_MELT_POOL_WIDTH]
            if not np.isnan(row[ColumnNames.MICRO_MELT_POOL_WIDTH])
            else MicrostructureInput.DEFAULT_MELT_POOL_WIDTH,
            melt_pool_depth=row[ColumnNames.MICRO_MELT_POOL_DEPTH]
            if not np.isnan(row[ColumnNames.MICRO_MELT_POOL_DEPTH])
            else MicrostructureInput.DEFAULT_MELT_POOL_DEPTH,
            random_seed=row[ColumnNames.RANDOM_SEED]
            if not np.isnan(row[ColumnNames.RANDOM_SEED])
            else MicrostructureInput.DEFAULT_RANDOM_SEED,
        )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides definitions for common controls on parametric study plots."""

import pandas as pd
import panel as pn

from ansys.additive.core.parametric_study import ColumnNames


def _common_controls(df: pd.DataFrame):
    ht_options = {
        k: v for (k, v) in [(f"{t:.2f} °C", t) for t in df[ColumnNames.HEATER_TEMPERATURE].unique()]
    }
    ht_select = pn.widgets.Select(
        name="Heater Temperature", options=ht_options, sizing_mode="stretch_width"
    ).servable()
    lt_options = {
        k: v
        for (k, v) in [(f"{t*1e6:.0f} µm", t) for t in df[ColumnNames.LAYER_THICKNESS].unique()]
    }
    lt_select = pn.widgets.Select(
        name="Layer Thickness", options=lt_options, sizing_mode="stretch_width"
    ).servable()
    bd_options = {
        k: v for (k, v) in [(f"{t*1e6:.0f} µm", t) for t in df[ColumnNames.BEAM_DIAMETER].unique()]
    }
    bd_select = pn.widgets.Select(
        name="Beam Diameter", options=bd_options, sizing_mode="stretch_width"
    ).servable()
    sa_options = {
        k: v for (k, v) in [(f"{t:.1f} °", t) for t in df[ColumnNames.START_ANGLE].unique()]
    }
    sa_select = pn.widgets.Select(
        name="Start Angle", options=sa_options, sizing_mode="stretch_width"
    ).servable()
    ra_options = {
        k: v for (k, v) in [(f"{t:.1f} °", t) for t in df[ColumnNames.ROTATION_ANGLE].unique()]
    }
    ra_select = pn.widgets.Select(
        name="Rotation Angle", options=ra_options, sizing_mode="stretch_width"
    ).servable()
    hs_options = {
        k: v for (k, v) in [(f"{t*1e6:.0f} µm", t) for t in df[ColumnNames.HATCH_SPACING].unique()]
    }
    hs_select = pn.widgets.Select(
        name="Hatch Spacing", options=hs_options, sizing_mode="stretch_width"
    ).servable()
    sw_options = {
        k: v for (k, v) in [(f"{t*1e3:.0f} mm", t) for t in df[ColumnNames.STRIPE_WIDTH].unique()]
    }
    sw_select = pn.widgets.Select(
        name="Stripe Width", options=sw_options, sizing_mode="stretch_width"
    ).servable()
    return (
        ht_select,
        lt_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
    )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides an interactive interface for the parametric study simulation table."""
from __future__ import annotations

import os

import pandas as pd
import panel as pn

from ansys.additive.core import SimulationStatus, SimulationType
from ansys.additive.core.misc import short_uuid
from ansys.additive.core.parametric_study import ColumnNames, ParametricStudy

# global variables
_df = None
_ps = None
_iter_select = None
_pri_select = None

pn.extension("tabulator")


def show_table(ps: ParametricStudy, page_size: int = 10):
    """Generate an interactive display of the parametric study table.

    Parameters
    ----------
    ps : :class:`ParametricStudy <ansys.additive.core.parametric_study.ParametricStudy>`
        Parametric study to display.
    page_size : int, 10
        Number of table rows to display per page.

    Returns
    -------
    :class:`panel.Column <panel.Column>`
        Interactive table.
    """
    global _df, _ps, _iter_select, _pri_select

    _ps = ps
    _df = _ps.data_frame()

    (
        _iter_select,
        _pri_select,
        type_select,
        status_select,
    ) = __init_controls(_df)

    table = pn.widgets.Tabulator(
        _df,
        pagination="local",
        page_size=page_size,
        layout="fit_data_stretch",
        editors=__editors(_df),
    ).servable()
    table.add_filter(_iter_select, ColumnNames.ITERATION)
    table.add_filter(_pri_select, ColumnNames.PRIORITY)
    table.add_filter(type_select, ColumnNames.TYPE)
    table.add_filter(status_select, ColumnNames.STATUS)
    table.on_edit(__on_edit)

    control_bar = pn.Row(
        _iter_select,
        _pri_select,
        type_select,
        status_select,
        sizing_mode="stretch_width",
    )
    col = pn.Column(
        pn.widgets.StaticText(value="<b>Parametric Study</b>"),
        control_bar,
        pn.widgets.StaticText(value="<i>Use CTRL + click to select multiple values.</i>"),
        table,
        sizing_mode="stretch_both",
    ).servable()
    if os.getenv("GENERATING_DOCS"):
        name = show_table.__name__
        col.save(f"{name}_{short_uuid()}.png")
        col.__repr__ = lambda: name
    return col


def __init_controls(df: pd.DataFrame) -> tuple[pn.widgets.MultiSelect, ...]:
    iter_options = sorted(df[ColumnNames.ITERATION].unique().tolist())
    pri_options = sorted(df[ColumnNames.PRIORITY].unique().tolist())
    type_options = [
        SimulationType.SINGLE_BEAD,
        SimulationType.POROSITY,
        SimulationType.MICROSTRUCTURE,
    ]
    status_options = [
        SimulationStatus.PENDING,
        SimulationStatus.COMPLETED,
        SimulationStatus.SKIP,
        SimulationStatus.ERROR,
    ]
    iter_select = pn.widgets.MultiSelect(
        name="Iteration", options=iter_options, value=iter_options, sizing_mode="stretch_width"
    ).servable()
    pri_select = pn.widgets.MultiSelect(
        name="Priority", options=pri_options, value=pri_options, sizing_mode="stretch_width"
    ).servable()
    type_select = pn.widgets.MultiSelect(
        name="Type", options=type_options, value=type_options, sizing_mode="stretch_width"
    ).servable()
    status_select = pn.widgets.MultiSelect(
        name="Status", options=status_options, value=status_options, sizing_mode="stretch_width"
    ).servable()

    return iter_select, pri_select, type_select, status_select


def __editors(df: pd.DataFrame):
    editors = {}
    for col in df.columns:
        if col == ColumnNames.STATUS:
            editors[col] = {
                "type": "list",
                "values": [
                    SimulationStatus.PENDING,
                    SimulationStatus.COMPLETED,
                    SimulationStatus.ERROR,
                    SimulationStatus.SKIP,
                ],
                "multiselect": False,
            }
        elif col in [ColumnNames.PRIORITY, ColumnNames.ITERATION]:
            editors[col] = {
                "type": "number",
                "step": 1,
            }
        else:
            editors[col] = None
    return editors


def __on_edit(event: any):
    global _df, _iter_select, _pri_select
    if event.column in [ColumnNames.STATUS, ColumnNames.PRIORITY, ColumnNames.ITERATION]:
        id = _df.loc[event.row, ColumnNames.ID]
        if event.column == ColumnNames.STATUS:
            _ps.set_status(id, event.value)
        elif event.column == ColumnNames.PRIORITY:
            _ps.set_priority(id, event.value)
            _pri_select.options = sorted(_ps.data_frame()[ColumnNames.PRIORITY].unique().tolist())
            # _pri_select.value is not type(list), so we need to create a list
            selections = [x for x in _pri_select.value]
            selections.append(event.value)
            _pri_select.value = selections
        elif event.column == ColumnNames.ITERATION:
            _ps.set_iteration(id, event.value)
            _iter_select.options = sorted(_ps.data_frame()[ColumnNames.ITERATION].unique().tolist())
            # _iter_select.value is not type(list), so we need to create a list
            selections = [x for x in _iter_select.value]
            selections.append(event.value)
            _iter_select.value = selections
        _df = _ps.data_frame()



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides input and result summary containers for single bead simulations."""
import math

from ansys.api.additive.v0.additive_domain_pb2 import MeltPool as MeltPoolMessage
from ansys.api.additive.v0.additive_domain_pb2 import SingleBeadInput as SingleBeadInputMessage
from ansys.api.additive.v0.additive_simulation_pb2 import SimulationRequest
from pandas import DataFrame

from ansys.additive.core.machine import AdditiveMachine
from ansys.additive.core.material import AdditiveMaterial


class SingleBeadInput:
    """Provides input parameters for a single bead simulation."""

    DEFAULT_BEAD_LENGTH = 3e-3
    """Default bead length (m)."""
    MIN_BEAD_LENGTH = 1e-3
    """Minimum bead length (m)."""
    MAX_BEAD_LENGTH = 1e-2
    """Maximum bead length (m)."""

    def __init__(
        self,
        id: str = "",
        bead_length: float = DEFAULT_BEAD_LENGTH,
        machine: AdditiveMachine = AdditiveMachine(),
        material: AdditiveMaterial = AdditiveMaterial(),
    ):
        """Initialize a ``SingleBeadInput`` object."""
        self.id = id
        self.bead_length = bead_length
        self.machine = machine
        self.material = material

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            if k == "_machine" or k == "_material":
                repr += "\n" + k.replace("_", "", 1) + ": " + str(getattr(self, k))
            else:
                repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, SingleBeadInput):
            return False
        return (
            self.id == __o.id
            and self.bead_length == __o.bead_length
            and self.machine == __o.machine
            and self.material == __o.material
        )

    def __validate_range(self, value, min, max, name):
        if math.isnan(value):
            raise ValueError("{} must be a number.".format(name))
        if value < min or value > max:
            raise ValueError("{} must be between {} and {}.".format(name, min, max))

    @property
    def id(self) -> str:
        """User-provided ID for the simulation."""
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def machine(self) -> AdditiveMachine:
        """Machine parameters."""
        return self._machine

    @machine.setter
    def machine(self, value):
        self._machine = value

    @property
    def material(self) -> AdditiveMaterial:
        """Material parameters."""
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def bead_length(self) -> float:
        """Length (m) of bead to simulate.

        Valid values are from the :obj:`MIN_BEAD_LENGTH` value to the
        :obj:`MAX_BEAD_LENGTH` value.
        """
        return self._bead_length

    @bead_length.setter
    def bead_length(self, value):
        self.__validate_range(value, self.MIN_BEAD_LENGTH, self.MAX_BEAD_LENGTH, "bead_length")
        self._bead_length = value

    def _to_simulation_request(self) -> SimulationRequest:
        """Convert this object into a simulation request message."""
        input = SingleBeadInputMessage(
            machine=self.machine._to_machine_message(),
            material=self.material._to_material_message(),
            bead_length=self.bead_length,
        )
        return SimulationRequest(id=self.id, single_bead_input=input)


class MeltPoolColumnNames:
    """Provides column names for the melt pool data frame."""

    WIDTH = "width"
    """Width of melt pool (m)."""
    DEPTH = "depth"
    """Depth of melt pool (m)."""
    LENGTH = "length"
    """Length of melt pool (m)."""
    REFERENCE_WIDTH = "reference_width"
    """Width of melt pool at the surface of the base plate (m)."""
    REFERENCE_DEPTH = "reference_depth"
    """Depth of melt pool measured from the surface of the base plate (m)."""


class MeltPool:
    """Contains the melt pool size dimensions for each time step during a single bead simulation."""

    def __init__(self, msg: MeltPoolMessage):
        """Initialize a ``MeltPool`` object."""
        bead_length = [ts.laser_x for ts in msg.time_steps]
        length = [ts.length for ts in msg.time_steps]
        width = [ts.width for ts in msg.time_steps]
        depth = [ts.depth for ts in msg.time_steps]
        reference_width = [ts.reference_width for ts in msg.time_steps]
        reference_depth = [ts.reference_depth for ts in msg.time_steps]
        self._df = DataFrame(
            index=bead_length,
            data={
                MeltPoolColumnNames.LENGTH: length,
                MeltPoolColumnNames.WIDTH: width,
                MeltPoolColumnNames.DEPTH: depth,
                MeltPoolColumnNames.REFERENCE_WIDTH: reference_width,
                MeltPoolColumnNames.REFERENCE_DEPTH: reference_depth,
            },
        )
        self._df.index.name = "bead_length"

    def data_frame(self) -> DataFrame:
        """Get the data frame containing the melt pool data.

        Values are in meters.

        Indices:
            - ``bead_length``: Length of the bead at each time step.

        Columns:
            - :obj:`MeltPoolColumnNames.LENGTH`.
            - :obj:`MeltPoolColumnNames.WIDTH`.
            - :obj:`MeltPoolColumnNames.DEPTH`.
            - :obj:`MeltPoolColumnNames.REFERENCE_WIDTH`.
            - :obj:`MeltPoolColumnNames.REFERENCE_DEPTH`.
        """
        return self._df.copy()

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, MeltPool):
            return False
        return self._df.equals(__o._df)

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        repr += self._df.to_string()
        return repr


class SingleBeadSummary:
    """Provides a summary of a single bead simulation."""

    def __init__(
        self,
        input: SingleBeadInput,
        msg: MeltPoolMessage,
    ):
        """Initialize a ``SingleBeadSummary`` object."""
        if not isinstance(input, SingleBeadInput):
            raise ValueError("Invalid input type passed to init, " + self.__class__.__name__)
        if not isinstance(msg, MeltPoolMessage):
            raise ValueError("Invalid message type passed to init, " + self.__class__.__name__)
        self._input = input
        self._melt_pool = MeltPool(msg)

    @property
    def input(self) -> SingleBeadInput:
        """Simulation input."""
        return self._input

    @property
    def melt_pool(self) -> MeltPool:
        """Resulting melt pool."""
        return self._melt_pool

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        for k in self.__dict__:
            repr += k.replace("_", "", 1) + ": " + str(getattr(self, k)) + "\n"
        return repr



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a client for interacting with the Additive service."""

from __future__ import annotations

from collections.abc import Iterator
import concurrent.futures
from datetime import datetime
import hashlib
import logging
import os
import zipfile

from ansys.api.additive import __version__ as api_version
from ansys.api.additive.v0.additive_domain_pb2 import ProgressState
from ansys.api.additive.v0.additive_materials_pb2 import GetMaterialRequest
from ansys.api.additive.v0.additive_simulation_pb2 import UploadFileRequest
from google.protobuf.empty_pb2 import Empty
import grpc

from ansys.additive.core import USER_DATA_PATH, __version__
from ansys.additive.core.download import download_file
from ansys.additive.core.exceptions import BetaFeatureNotEnabledError
from ansys.additive.core.material import AdditiveMaterial
from ansys.additive.core.material_tuning import MaterialTuningInput, MaterialTuningSummary
from ansys.additive.core.microstructure import MicrostructureInput, MicrostructureSummary
from ansys.additive.core.microstructure_3d import Microstructure3DInput, Microstructure3DSummary
import ansys.additive.core.misc as misc
from ansys.additive.core.porosity import PorosityInput, PorositySummary
from ansys.additive.core.progress_logger import ProgressLogger
from ansys.additive.core.server_connection import DEFAULT_PRODUCT_VERSION, ServerConnection
from ansys.additive.core.simulation import SimulationError
from ansys.additive.core.single_bead import SingleBeadInput, SingleBeadSummary
from ansys.additive.core.thermal_history import ThermalHistoryInput, ThermalHistorySummary


class Additive:
    """Provides the client interface to one or more Additive services.

    In a typical cloud environment, a single Additive service with load balancing and
    auto-scaling is used. The ``Additive`` client connects to the service via a
    single connection. However, for atypical environments or when running on localhost,
    the ``Additive`` client can perform crude load balancing by connecting to multiple
    servers and distributing simulations across them. You can use the ``server_connections``,
    ``nservers``, and ``nsims_per_server`` parameters to control the
    number of servers to connect to and the number of simulations to run on each
    server.

    Parameters
    ----------
    server_connections: list[str, grpc.Channel], None
        List of connection definitions for servers. The list may be a combination of strings and
        connected :class:`grpc.Channel <grpc.Channel>` objects. Strings use the format
        ``host:port`` to specify the server IPv4 address.
    host: str, default: None
        Host name or IPv4 address of the server. This parameter is ignored if the
        ``server_channels`` or ``channel`` parameters is other than ``None``.
    port: int, default: 50052
        Port number to use when connecting to the server.
    nsims_per_server: int, default: 1
        Number of simultaneous simulations to run on each server. Each simulation
        requires a license checkout. If a license is not available, the simulation
        fails.
    nservers: int, default: 1
        Number of Additive servers to start and connect to. This parameter is only
        applicable in `PyPIM`_-enabled cloud environments and on localhost. For
        this to work on localhost, the Additive portion of the Ansys Structures
        package must be installed. This parameter is ignored if the ``server_connections``
        parameter or ``host`` parameter is other than ``None``.
    product_version: str
        Version of the Ansys product installation in the form ``"YYR"``, where ``YY``
        is the two-digit year and ``R`` is the release number. For example, the release
        2024 R1 would be specified as ``241``. This parameter is only applicable in
        `PyPIM`_-enabled cloud environments and on localhost. Using an empty string
        or ``None`` uses the default product version.
    log_level: str, default: "INFO"
        Minimum severity level of messages to log.
    log_file: str, default: ""
        File name to write log messages to.
    enable_beta_features: bool, default: False
        Flag indicating if beta features are enabled.

    Examples
    --------
    Connect to a list of servers. Multiple connections to the same host are permitted.

    >>> additive = Additive(server_connections=["localhost:50052", "localhost:50052", "myserver:50052"])

    Connect to a single server using the host name and port number.

    >>> additive = Additive(host="additive.ansys.com", port=12345)

    Start and connect to two servers on localhost or in a
    `PyPIM`_-enabled cloud environment. Allow each server to run two
    simultaneous simulations.

    >>> additive = Additive(nsims_per_server=2, nservers=2)

    Start a single server on localhost or in a `PyPIM`_-enabled cloud environment.
    Use version 2024 R1 of the Ansys product installation.

    >>> additive = Additive(product_version="241")

    .. _PyPIM: https://pypim.docs.pyansys.com/version/stable/index.html
    """

    DEFAULT_ADDITIVE_SERVICE_PORT = 50052

    def __init__(
        self,
        server_connections: list[str | grpc.Channel] = None,
        host: str | None = None,
        port: int = DEFAULT_ADDITIVE_SERVICE_PORT,
        nsims_per_server: int = 1,
        nservers: int = 1,
        product_version: str = DEFAULT_PRODUCT_VERSION,
        log_level: str = "INFO",
        log_file: str = "",
        enable_beta_features: bool = False,
    ) -> None:
        """Initialize server connections."""
        if product_version is None or product_version == "":
            product_version = DEFAULT_PRODUCT_VERSION

        self._log = Additive._create_logger(log_file, log_level)
        self._log.debug("Logging set to %s", log_level)

        self._servers = Additive._connect_to_servers(
            server_connections, host, port, nservers, product_version, self._log
        )
        self._nsims_per_server = nsims_per_server
        self._enable_beta_features = enable_beta_features

        # Setup data directory
        self._user_data_path = USER_DATA_PATH
        if not os.path.exists(self._user_data_path):  # pragma: no cover
            os.makedirs(self._user_data_path)
        print("user data path: " + self._user_data_path)

    @staticmethod
    def _create_logger(log_file, log_level) -> logging.Logger:
        """Instantiate the logger."""
        format = "%(asctime)s %(message)s"
        datefmt = "%Y-%m-%d %H:%M:%S"
        numeric_level = getattr(logging, log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError("Invalid log level: %s" % log_level)
        logging.basicConfig(
            level=numeric_level,
            format=format,
            datefmt=datefmt,
        )
        log = logging.getLogger(__name__)
        if log_file:
            file_handler = logging.FileHandler(str(log_file))
            file_handler.setLevel(numeric_level)
            file_handler.setFormatter(logging.Formatter(format))
            log.file_handler = file_handler
            log.addHandler(file_handler)
        return log

    @staticmethod
    def _connect_to_servers(
        server_connections: list[str | grpc.Channel] = None,
        host: str | None = None,
        port: int = DEFAULT_ADDITIVE_SERVICE_PORT,
        nservers: int = 1,
        product_version: str = DEFAULT_PRODUCT_VERSION,
        log: logging.Logger = None,
    ) -> list[ServerConnection]:
        """Connect to Additive servers, starting them if necessary."""
        connections = []
        if server_connections:
            for target in server_connections:
                if isinstance(target, grpc.Channel):
                    connections.append(ServerConnection(channel=target, log=log))
                else:
                    connections.append(ServerConnection(addr=target, log=log))
        elif host:
            connections.append(ServerConnection(addr=f"{host}:{port}", log=log))
        elif os.getenv("ANSYS_ADDITIVE_ADDRESS"):
            connections.append(ServerConnection(addr=os.getenv("ANSYS_ADDITIVE_ADDRESS"), log=log))
        else:
            for _ in range(nservers):
                connections.append(ServerConnection(product_version=product_version, log=log))

        return connections

    @property
    def nsims_per_server(self) -> int:
        """Number of simultaneous simulations to run on each server."""
        return self._nsims_per_server

    @nsims_per_server.setter
    def nsims_per_server(self, value: int) -> None:
        """Set the number of simultaneous simulations to run on each server."""
        if value < 1:
            raise ValueError("Number of simulations per server must be greater than zero.")
        self._nsims_per_server = value

    @property
    def enable_beta_features(self) -> bool:
        """Flag indicating if beta features are enabled."""
        return self._enable_beta_features

    @enable_beta_features.setter
    def enable_beta_features(self, value: bool) -> None:
        """Set the flag indicating if beta features are enabled."""
        self._enable_beta_features = value

    def about(self) -> None:
        """Print information about the client and server."""
        print(f"Client {__version__}, API version: {api_version}")
        if self._servers is None:
            print("Client is not connected to a server.")
            return
        else:
            for server in self._servers:
                print(server.status())

    def simulate(
        self,
        inputs: (
            SingleBeadInput
            | PorosityInput
            | MicrostructureInput
            | ThermalHistoryInput
            | Microstructure3DInput
            | list
        ),
    ) -> (
        SingleBeadSummary
        | PorositySummary
        | MicrostructureSummary
        | ThermalHistorySummary
        | Microstructure3DSummary
        | SimulationError
        | list
    ):
        """Execute additive simulations.

        Parameters
        ----------
        inputs: SingleBeadInput, PorosityInput, MicrostructureInput, ThermalHistoryInput,
        Microstructure3DInput, list
            Parameters to use for simulations. A list of inputs may be provided to execute multiple
            simulations.

        Returns
        -------
        SingleBeadSummary, PorositySummary, MicrostructureSummary, ThermalHistorySummary,
        Microstructure3DSummary, SimulationError, list
            One or more summaries of simulation results. If a list of inputs is provided, a
            list is returned.
        """
        self._check_for_duplicate_id(inputs)

        if not isinstance(inputs, list):
            return self._simulate(inputs, self._servers[0], show_progress=True)

        summaries = []
        print(
            f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} Completed 0 of {len(inputs)} simulations",
            end="",
        )
        threads = min(len(inputs), len(self._servers) * self._nsims_per_server)
        with concurrent.futures.ThreadPoolExecutor(threads) as executor:
            futures = []
            for i, input in enumerate(inputs):
                server_id = i % len(self._servers)
                futures.append(
                    executor.submit(
                        self._simulate,
                        input=input,
                        server=self._servers[server_id],
                        show_progress=False,
                    )
                )
            for future in concurrent.futures.as_completed(futures):
                summary = future.result()
                if isinstance(summary, SimulationError):
                    print(f"\nError: {summary.message}")
                summaries.append(summary)
                timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                print(
                    f"\r{timestamp} Completed {len(summaries)} of {len(inputs)} simulations",
                    end="",
                )
        print("")
        return summaries

    def _simulate(
        self,
        input: (
            SingleBeadInput
            | PorosityInput
            | MicrostructureInput
            | ThermalHistoryInput
            | Microstructure3DInput
        ),
        server: ServerConnection,
        show_progress: bool = False,
    ) -> (
        SingleBeadSummary
        | PorositySummary
        | MicrostructureSummary
        | ThermalHistorySummary
        | Microstructure3DSummary
        | SimulationError
    ):
        """Execute a single simulation.

        Parameters
        ----------
        input: SingleBeadInput, PorosityInput, MicrostructureInput, ThermalHistoryInput,
        Microstructure3DInput
            Parameters to use for simulation.

        server: ServerConnection
            Server to use for the simulation.

        show_progress: bool, False
            Whether to send progress updates to the user interface.

        Returns
        -------
        SingleBeadSummary, PorositySummary, MicrostructureSummary, ThermalHistorySummary,
        Microstructure3DSummary, SimulationError
        """
        logger = None
        if show_progress:
            logger = ProgressLogger("Simulation")  # pragma: no cover

        if input.material == AdditiveMaterial():
            raise ValueError("A material is not assigned to the simulation input")

        if isinstance(input, Microstructure3DInput) and self.enable_beta_features is False:
            raise BetaFeatureNotEnabledError(
                "3D microstructure simulations require beta features to be enabled.\n"
                + "Set enable_beta_features=True when creating the Additive client."
            )

        try:
            request = None
            if isinstance(input, ThermalHistoryInput):
                return self._simulate_thermal_history(input, USER_DATA_PATH, server, logger)
            else:
                request = input._to_simulation_request()

            for response in server.simulation_stub.Simulate(request):
                if response.HasField("progress"):
                    if logger:
                        logger.log_progress(response.progress)  # pragma: no cover
                    if response.progress.state == ProgressState.PROGRESS_STATE_ERROR:
                        raise Exception(response.progress.message)
                if response.HasField("melt_pool"):
                    return SingleBeadSummary(input, response.melt_pool)
                if response.HasField("porosity_result"):
                    return PorositySummary(input, response.porosity_result)
                if response.HasField("microstructure_result"):
                    return MicrostructureSummary(
                        input, response.microstructure_result, self._user_data_path
                    )
                if response.HasField("microstructure_3d_result"):
                    return Microstructure3DSummary(
                        input, response.microstructure_3d_result, self._user_data_path
                    )
        except Exception as e:
            return SimulationError(input, str(e))

    def materials_list(self) -> list[str]:
        """Get a list of material names used in additive simulations.

        Returns
        -------
        list[str]
            Names of available additive materials.
        """
        response = self._servers[0].materials_stub.GetMaterialsList(Empty())
        return [n for n in response.names]

    def material(self, name: str) -> AdditiveMaterial:
        """Get a material for use in an additive simulation.

        Parameters
        ----------

        name: str
            Name of material.

        Returns
        -------
        AdditiveMaterial
        """
        request = GetMaterialRequest(name=name)
        result = self._servers[0].materials_stub.GetMaterial(request)
        return AdditiveMaterial._from_material_message(result)

    @staticmethod
    def load_material(
        parameters_file: str, thermal_lookup_file: str, characteristic_width_lookup_file: str
    ) -> AdditiveMaterial:
        """Load a user-provided material definition.

        Parameters
        ----------
        parameters_file: str
            Name of the JSON file containing material parameters. For more information, see
            `Create Material Configuration File (.json)
            <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_tool_create_tables.html>`_
            in the *Additive Manufacturing Beta Features* documentation.
        thermal_lookup_file: str
            Name of the CSV file containing the lookup table for thermal dependent properties.
            For more information, see `Create Material Lookup File (.csv)
            <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_create_mat_lookup.html>`_
            in the *Additive Manufacturing Beta Features* documentation.
        characteristic_width_lookup_file: str
            Name of the CSV file containing the lookup table for characteristic melt pool width. For
            more information, see
            `Find Characteristic Width Values and Generate Characteristic Width File (.csv)
            <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_tool_find_cw.html>`_
            in the *Additive Manufacturing Beta Features* documentation.
        """
        material = AdditiveMaterial()
        material._load_parameters(parameters_file)
        material._load_thermal_properties(thermal_lookup_file)
        material._load_characteristic_width(characteristic_width_lookup_file)
        return material

    def tune_material(
        self, input: MaterialTuningInput, out_dir: str = USER_DATA_PATH
    ) -> MaterialTuningSummary:
        """Tune a custom material for use with additive simulations.

        This method performs the same function as the Material Tuning Tool
        described in
        `Find Simulation Parameters to Match Simulation to Experiments
        <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_print_udm_tool_match_sim_to_exp.html>`_.
        It is used for one step in the material tuning process. The other steps
        are described in
        `Chapter 2: Material Tuning Tool (Beta) to Create User Defined Materials
        <https://ansyshelp.ansys.com/account/secured?returnurl=/Views/Secured/corp/v232/en/add_beta/add_science_BETA_material_tuning_tool.html>`_.

        Parameters
        ----------
        input: MaterialTuningInput
            Input parameters for material tuning.

        Returns
        -------
        MaterialTuningSummary
            Summary of material tuning.
        """  # noqa: E501
        if input.id == "":
            input.id = misc.short_uuid()
        if out_dir == USER_DATA_PATH:
            out_dir = os.path.join(out_dir, input.id)

        if os.path.exists(out_dir):
            raise ValueError(
                f"Directory {out_dir} already exists. Delete or choose a different output directory."
            )

        request = input._to_request()

        for response in self._servers[0].materials_stub.TuneMaterial(request):
            if response.HasField("progress"):
                if response.progress.state == ProgressState.PROGRESS_STATE_ERROR:
                    raise Exception(response.progress.message)
                else:
                    for m in response.progress.message.splitlines():
                        if (
                            "License successfully" in m
                            or "Starting ThermalSolver" in m
                            or "threads for solver" in m
                        ):
                            continue
                        print(m)
            if response.HasField("result"):
                return MaterialTuningSummary(input, response.result, out_dir)

    def __file_upload_reader(
        self, file_name: str, chunk_size=2 * 1024**2
    ) -> Iterator[UploadFileRequest]:
        """Read a file and return an iterator of UploadFileRequests."""
        file_size = os.path.getsize(file_name)
        short_name = os.path.basename(file_name)
        with open(file_name, mode="rb") as f:
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                yield UploadFileRequest(
                    name=short_name,
                    total_size=file_size,
                    content=chunk,
                    content_md5=hashlib.md5(chunk).hexdigest(),
                )

    def _simulate_thermal_history(
        self,
        input: ThermalHistoryInput,
        out_dir: str,
        server: ServerConnection,
        logger: ProgressLogger | None = None,
    ) -> ThermalHistorySummary:
        """Execute a thermal history simulation.

        Parameters
        ----------
        input: ThermalHistoryInput
            Simulation input parameters.
        out_dir: str
            Folder path for output files.
        server: ServerConnection
            Server to use for the simulation.
        logger: ProgressLogger
            Log message handler.

        Returns
        -------
        :class:`ThermalHistorySummary`
        """
        if input.geometry is None or input.geometry.path == "":
            raise ValueError("The geometry path is not defined in the simulation input")

        remote_geometry_path = ""
        for response in server.simulation_stub.UploadFile(
            self.__file_upload_reader(input.geometry.path)
        ):
            remote_geometry_path = response.remote_file_name
            if logger:
                logger.log_progress(response.progress)  # pragma: no cover
            if response.progress.state == ProgressState.PROGRESS_STATE_ERROR:
                raise Exception(response.progress.message)

        request = input._to_simulation_request(remote_geometry_path=remote_geometry_path)
        for response in server.simulation_stub.Simulate(request):
            if response.HasField("progress"):
                if logger:
                    logger.log_progress(response.progress)  # pragma: no cover
                if (
                    response.progress.state == ProgressState.PROGRESS_STATE_ERROR
                    and "WARN" not in response.progress.message
                ):
                    raise Exception(response.progress.message)
            if response.HasField("thermal_history_result"):
                path = os.path.join(out_dir, input.id, "coax_ave_output")
                local_zip = download_file(
                    server.simulation_stub,
                    response.thermal_history_result.coax_ave_zip_file,
                    path,
                )
                with zipfile.ZipFile(local_zip, "r") as zip:
                    zip.extractall(path)
                os.remove(local_zip)
                return ThermalHistorySummary(input, path)

    def _check_for_duplicate_id(self, inputs):
        if not isinstance(inputs, list):
            return
        ids = []
        for input in inputs:
            if input.id == "":
                input.id = misc.short_uuid()
            if input.id in ids:
                raise ValueError(f'Duplicate simulation ID "{input.id}" in input list')
            ids.append(input.id)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a function for downloading files from the server to the client."""
import hashlib
import os

from ansys.api.additive.v0.additive_simulation_pb2 import DownloadFileRequest
from ansys.api.additive.v0.additive_simulation_pb2_grpc import SimulationServiceStub

from ansys.additive.core.progress_logger import Progress, ProgressLogger, ProgressState


def download_file(
    stub: SimulationServiceStub,
    remote_file_name: str,
    local_folder: str,
    logger: ProgressLogger = None,
) -> str:
    """Download a file from the server to the localhost.

    Parameters
    ----------
    remote_file_name: str
        Path to file on the server.
    local_folder: str
        Folder on your localhost to write your file to.
    logger: ProgressLogger
        Log message handler.

    Returns
    -------
    str
        Local path of downloaded file.
    """

    if not os.path.isdir(local_folder):
        os.makedirs(local_folder)

    dest = os.path.join(local_folder, os.path.basename(remote_file_name))
    request = DownloadFileRequest(remote_file_name=remote_file_name)

    with open(dest, "wb") as f:
        for response in stub.DownloadFile(request):
            if logger:
                logger.log_progress(response.progress)  # pragma: no cover
            if len(response.content) > 0:
                md5 = hashlib.md5(response.content).hexdigest()
                if md5 != response.content_md5:
                    if logger:  # pragma: no cover
                        logger.log_progress(
                            Progress(
                                state=ProgressState.PROGRESS_STATE_ERROR,
                                message="Download error, MD5 sums did not match",
                            )
                        )
                    raise ValueError("Download error, MD5 sums did not match")
                f.write(response.content)
    return dest



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides interactive plot of porosity results."""

from __future__ import annotations

import math
import os

import numpy as np
import pandas as pd
import panel as pn
import plotly.graph_objects as go

from ansys.additive.core import SimulationStatus, SimulationType
from ansys.additive.core.misc import short_uuid
from ansys.additive.core.parametric_study import ColumnNames, ParametricStudy

from ._common_controls import _common_controls

# global variables
_range_slider = None

pn.extension("plotly")


def porosity_eval_plot(ps: ParametricStudy):
    """Generate a heat map plot of porosity results to determine parametric regions with desirable relative density statistics.

    Parameters
    ----------
    ps : ParametricStudy to plot.

    Returns
    -------
    :class: `panel.Row <panel.Row>`
        Interactive plot.
    """  # noqa
    df = __data_frame(ps)
    (
        ht_select,
        lt_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
        _range_slider,
    ) = __init_controls(df)
    side_bar = pn.Column(
        pn.Spacer(height=50),
        _range_slider,
        lt_select,
        ht_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
        width=200,
    )
    plot_view = pn.bind(
        __update_plot,
        df,
        ht_select,
        lt_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
        _range_slider,
    )
    plot = pn.Row(
        side_bar,
        pn.pane.Plotly(plot_view, sizing_mode="stretch_both", min_height=600),
        sizing_mode="stretch_both",
    ).servable()
    if os.getenv("GENERATING_DOCS"):
        name = porosity_eval_plot.__name__
        plot.save(f"{name}_{short_uuid()}.png")
        plot.__repr__ = lambda: name
    return plot


def __data_frame(ps: ParametricStudy) -> pd.DataFrame:
    df = ps.data_frame()
    df = df[
        (df[ColumnNames.TYPE] == SimulationType.POROSITY)
        & (df[ColumnNames.STATUS] == SimulationStatus.COMPLETED)
    ]
    df.update(
        df[
            [
                ColumnNames.RELATIVE_DENSITY,
            ]
        ].fillna(0)
    )
    return df


def __init_controls(df: pd.DataFrame):
    global _range_slider

    (
        ht_select,
        lt_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
    ) = _common_controls(df)
    range_end = 1.0
    _range_slider = pn.widgets.RangeSlider(
        name="Relative density",
        sizing_mode="stretch_width",
        start=0,
        end=range_end,
        value=(0.85 * range_end, 1.0 * range_end),
        step=0.01,
        bar_color="green",
    ).servable()
    return (
        ht_select,
        lt_select,
        bd_select,
        sa_select,
        ra_select,
        hs_select,
        sw_select,
        _range_slider,
    )


def __contour_colorscale() -> list:
    return [
        [0.0, "green"],
        [0.1, "green"],
        [0.15, "yellow"],
        [0.5, "firebrick"],
        [1.0, "firebrick"],
    ]


# @pn.cache
def __update_plot(
    df: pd.DataFrame,
    ht: float,
    lt: float,
    bd: float,
    sa: float,
    ra: float,
    hs: float,
    sw: float,
    range: tuple[float, float],
) -> go.Figure:
    fig = go.Figure()

    x, y, z = __contour_data(df, ht, lt, bd, sa, ra, hs, sw, range)
    contour = go.Heatmap(
        x=x,
        y=y,
        z=z,
        colorscale=__contour_colorscale(),
        showscale=False,
        connectgaps=True,
        hoverinfo="skip",
        zsmooth="best",
    )
    fig.add_trace(contour)

    scatter_x, scatter_y, z_scatter = __scatter_data(df, ht, lt, bd, sa, ra, hs, sw, range)
    scatter = go.Scatter(
        x=scatter_x,
        y=scatter_y,
        mode="markers+text",
        text=[f"<b>{z:.2f}</b>" for z in z_scatter],
        textposition="top center",
        marker=dict(color="black", size=5),
        cliponaxis=False,
    )
    fig.add_trace(scatter)
    title = "Porosity Relative density"
    fig.update_layout(title_text=title, plot_bgcolor="white")
    fig.update_xaxes(
        title_text="Scan Speed (m/s)",
        showgrid=True,
        gridwidth=1,
        gridcolor="lightgray",
        title_font=dict(size=12),
        range=[min(x) - 0.1, max(x) + 0.2],
    )
    fig.update_yaxes(
        title_text="Laser Power (W)",
        showgrid=True,
        gridwidth=1,
        gridcolor="lightgray",
        title_font=dict(size=12),
        range=[min(y) - 20, max(y) + 70],
    )

    return fig


def __contour_data(
    df: pd.DataFrame,
    ht: float,
    lt: float,
    bd: float,
    sa: float,
    ra: float,
    hs: float,
    sw: float,
    range: tuple[float, float],
) -> tuple[list, list, list]:
    """Get lists of scan speed, laser power, and relative density
    values."""

    idx = df[
        (df[ColumnNames.LAYER_THICKNESS] == lt)
        & (df[ColumnNames.BEAM_DIAMETER] == bd)
        & (df[ColumnNames.HEATER_TEMPERATURE] == ht)
        & (df[ColumnNames.START_ANGLE] == sa)
        & (df[ColumnNames.ROTATION_ANGLE] == ra)
        & (df[ColumnNames.HATCH_SPACING] == hs)
        & (df[ColumnNames.STRIPE_WIDTH] == sw)
    ].index

    df = df.loc[
        idx,
        [
            ColumnNames.LASER_POWER,
            ColumnNames.SCAN_SPEED,
            ColumnNames.RELATIVE_DENSITY,
        ],
    ]
    df.sort_values(
        by=[ColumnNames.LASER_POWER, ColumnNames.SCAN_SPEED],
        inplace=True,
    )
    speeds = df[ColumnNames.SCAN_SPEED].unique()
    powers = df[ColumnNames.LASER_POWER].unique()
    z_vals = []
    z_max = df[ColumnNames.RELATIVE_DENSITY].max()
    if math.isclose(z_max, 0, abs_tol=1e-5):
        z_max = 1
    min_range = range[0]
    max_range = range[1]
    for p in powers:
        row = []
        for v in speeds:
            if (
                len(
                    df.loc[
                        (df[ColumnNames.LASER_POWER] == p) & (df[ColumnNames.SCAN_SPEED] == v),
                        ColumnNames.RELATIVE_DENSITY,
                    ].values
                )
                > 0
            ):
                z = df.loc[
                    (df[ColumnNames.LASER_POWER] == p) & (df[ColumnNames.SCAN_SPEED] == v),
                    ColumnNames.RELATIVE_DENSITY,
                ].values[0]
                z = round(z, 2)
                if z >= min_range and z <= max_range:
                    row.append(0.01)
                else:
                    row.append(0.1 + min(abs(max_range - z) / z_max, abs(min_range - z) / z_max))
            else:
                row.append(np.nan)
        z_vals.append(row)
    return (speeds, powers, z_vals)


def __scatter_data(
    df: pd.DataFrame,
    ht: float,
    lt: float,
    bd: float,
    sa: float,
    ra: float,
    hs: float,
    sw: float,
    float,
) -> tuple[list, list, list]:
    idx = df[
        (df[ColumnNames.LAYER_THICKNESS] == lt)
        & (df[ColumnNames.HEATER_TEMPERATURE] == ht)
        & (df[ColumnNames.BEAM_DIAMETER] == bd)
        & (df[ColumnNames.START_ANGLE] == sa)
        & (df[ColumnNames.ROTATION_ANGLE] == ra)
        & (df[ColumnNames.HATCH_SPACING] == hs)
        & (df[ColumnNames.STRIPE_WIDTH] == sw)
        & (~df[ColumnNames.RELATIVE_DENSITY].isna())
    ].index
    scatter_x = df.loc[idx, ColumnNames.SCAN_SPEED].tolist()
    scatter_y = df.loc[idx, ColumnNames.LASER_POWER].tolist()
    scatter_z = df.loc[idx, ColumnNames.RELATIVE_DENSITY].tolist()
    return (
        scatter_x,
        scatter_y,
        scatter_z,
    )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a container for machine parameters."""
import math

from ansys.api.additive.v0.additive_domain_pb2 import MachineSettings as MachineMessage

import ansys.additive.core.conversions as conversions


class MachineConstants:
    """Provides constants for additive manufacturing machine settings."""

    DEFAULT_LASER_POWER = 195
    """Default laser power (W)."""
    MIN_LASER_POWER = 50
    """Minimum laser power (W)."""
    MAX_LASER_POWER = 700
    """Maximum laser power (W)."""
    DEFAULT_SCAN_SPEED = 1.0
    """Default scan speed (m/s)."""
    MIN_SCAN_SPEED = 0.35
    """Minimum scan speed (m/s)."""
    MAX_SCAN_SPEED = 2.5
    """Maximum scan speed (m/s)."""
    DEFAULT_HEATER_TEMP = 80
    """Default heater temperature (C)."""
    MIN_HEATER_TEMP = 20
    """Minimum heater temperature (C)."""
    MAX_HEATER_TEMP = 500
    """Maximum heater temperature (C)."""
    DEFAULT_LAYER_THICKNESS = 5e-5
    """Default layer thickness (m)."""
    MIN_LAYER_THICKNESS = 1e-5
    """Minimum layer thickness (m)."""
    MAX_LAYER_THICKNESS = 1e-4
    """Maximum layer thickness (m)."""
    DEFAULT_BEAM_DIAMETER = 1e-4
    """Default beam diameter (m)."""
    MIN_BEAM_DIAMETER = 2e-5
    """Minimum beam diameter (m)."""
    MAX_BEAM_DIAMETER = 1.4e-4
    """Maximum beam diameter (m)."""
    DEFAULT_STARTING_LAYER_ANGLE = 57
    """Default starting layer angle (degrees)."""
    MIN_STARTING_LAYER_ANGLE = 0
    """Minimum starting layer angle (degrees)."""
    MAX_STARTING_LAYER_ANGLE = 180
    """Maximum starting layer angle (degrees)."""
    DEFAULT_LAYER_ROTATION_ANGLE = 67
    """Default layer rotation angle (degrees)."""
    MIN_LAYER_ROTATION_ANGLE = 0
    """Minimum layer rotation angle (degrees)."""
    MAX_LAYER_ROTATION_ANGLE = 180
    """Maximum layer rotation angle (degrees)."""
    DEFAULT_HATCH_SPACING = 1e-4
    """Default hatch spacing (m)."""
    MIN_HATCH_SPACING = 6e-5
    """Minimum hatch spacing (m)."""
    MAX_HATCH_SPACING = 2e-4
    """Maximum hatch spacing (m)."""
    DEFAULT_SLICING_STRIPE_WIDTH = 0.01
    """Default slicing stripe width (m)."""
    MIN_SLICING_STRIPE_WIDTH = 0.001
    """Minimum slicing stripe width (m)."""
    MAX_SLICING_STRIPE_WIDTH = 0.1
    """Maximum slicing stripe width (m)."""


class AdditiveMachine:
    """Provides the additive manufacturing machine settings used during simulations.

    Units are SI (m, kg, s, or K) unless otherwise noted. Exceptions include angles,
    which are (degrees), and the heater temperature, which is (degrees) Celsius.
    """

    def __init__(
        self,
        *,
        laser_power: float = MachineConstants.DEFAULT_LASER_POWER,
        scan_speed: float = MachineConstants.DEFAULT_SCAN_SPEED,
        heater_temperature: float = MachineConstants.DEFAULT_HEATER_TEMP,
        layer_thickness: float = MachineConstants.DEFAULT_LAYER_THICKNESS,
        beam_diameter: float = MachineConstants.DEFAULT_BEAM_DIAMETER,
        starting_layer_angle: float = MachineConstants.DEFAULT_STARTING_LAYER_ANGLE,
        layer_rotation_angle: float = MachineConstants.DEFAULT_LAYER_ROTATION_ANGLE,
        hatch_spacing: float = MachineConstants.DEFAULT_HATCH_SPACING,
        slicing_stripe_width: float = MachineConstants.DEFAULT_SLICING_STRIPE_WIDTH,
    ):
        """Initialize an ``AdditiveMachine`` object."""
        self.laser_power = laser_power
        self.scan_speed = scan_speed
        self.heater_temperature = heater_temperature
        self.layer_thickness = layer_thickness
        self.beam_diameter = beam_diameter
        self.starting_layer_angle = starting_layer_angle
        self.layer_rotation_angle = layer_rotation_angle
        self.hatch_spacing = hatch_spacing
        self.slicing_stripe_width = slicing_stripe_width

    def __repr__(self):
        repr = type(self).__name__ + "\n"
        repr += "laser_power: {} W\n".format(self._laser_power)
        repr += "scan_speed: {} m/s\n".format(self._scan_speed)
        repr += "heater_temperature: {} °C\n".format(self._heater_temperature)
        repr += "layer_thickness: {} m\n".format(self._layer_thickness)
        repr += "beam_diameter: {} m\n".format(self._beam_diameter)
        repr += "starting_layer_angle: {} °\n".format(self._starting_layer_angle)
        repr += "layer_rotation_angle: {} °\n".format(self._layer_rotation_angle)
        repr += "hatch_spacing: {} m\n".format(self._hatch_spacing)
        repr += "slicing_stripe_width: {} m\n".format(self._slicing_stripe_width)
        return repr

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, AdditiveMachine):
            return False
        for k in self.__dict__:
            if getattr(self, k) != getattr(__o, k):
                return False
        return True

    def __validate_range(self, value, min, max, name):
        if math.isnan(value):
            raise ValueError("{} must be a number.".format(name))
        if value < min or value > max:
            raise ValueError("{} must be between {} and {}.".format(name, min, max))

    @property
    def laser_power(self) -> float:
        """Scanning laser power (W).

        Valid values are from 50 to 700 Watts.
        """
        return self._laser_power

    @laser_power.setter
    def laser_power(self, value: float):
        self.__validate_range(
            value, MachineConstants.MIN_LASER_POWER, MachineConstants.MAX_LASER_POWER, "laser_power"
        )
        self._laser_power = value

    @property
    def scan_speed(self) -> float:
        """Laser scanning speed (m/s).

        Valid values are from 0.35 to 2.5 m/s.
        """
        return self._scan_speed

    @scan_speed.setter
    def scan_speed(self, value: float):
        self.__validate_range(
            value, MachineConstants.MIN_SCAN_SPEED, MachineConstants.MAX_SCAN_SPEED, "scan_speed"
        )
        self._scan_speed = value

    @property
    def heater_temperature(self) -> float:
        """Temperature (°C) of the machine build chamber heater.

        Valid values are from 20 to 500 °C.
        """
        return self._heater_temperature

    @heater_temperature.setter
    def heater_temperature(self, value: float):
        self.__validate_range(
            value,
            MachineConstants.MIN_HEATER_TEMP,
            MachineConstants.MAX_HEATER_TEMP,
            "heater_temperature",
        )
        self._heater_temperature = value

    @property
    def layer_thickness(self) -> float:
        """Thickness (m) of the powder layer deposited with each pass of the
        recoater blade.

        Valid values are from 1e-5 to 1e-4 m (10 to 100 µm).
        """
        return self._layer_thickness

    @layer_thickness.setter
    def layer_thickness(self, value: float):
        self.__validate_range(
            value,
            MachineConstants.MIN_LAYER_THICKNESS,
            MachineConstants.MAX_LAYER_THICKNESS,
            "layer_thickness",
        )
        self._layer_thickness = value

    @property
    def beam_diameter(self) -> float:
        """Width (m) of the laser on the powder or substrate surface defined
        using the D4σ beam diameter definition.

        Usually this value is provided by the machine manufacturer. It is sometimes
        called the laser spot diameter.

        Valid values are from 2e-5 to 1.4e-4 m (20 and 140 µm).
        """
        return self._beam_diameter

    @beam_diameter.setter
    def beam_diameter(self, value: float):
        self.__validate_range(
            value,
            MachineConstants.MIN_BEAM_DIAMETER,
            MachineConstants.MAX_BEAM_DIAMETER,
            "beam_diameter",
        )
        self._beam_diameter = value

    @property
    def starting_layer_angle(self) -> float:
        """Angle (°) to scan the first layer at.

        The angle is measured counterclockwise from the X axis, such that a value of 90°
        results in scan lines parallel to the Y axis.

        Valid values are from 0 to 180°.
        """
        return self._starting_layer_angle

    @starting_layer_angle.setter
    def starting_layer_angle(self, value: float):
        self.__validate_range(
            value,
            MachineConstants.MIN_STARTING_LAYER_ANGLE,
            MachineConstants.MAX_STARTING_LAYER_ANGLE,
            "starting_layer_angle",
        )
        self._starting_layer_angle = value

    @property
    def layer_rotation_angle(self) -> float:
        """Angle (°) to change the scan vector orientation from layer to layer.

        Valid values are from 0 to 180°.
        """
        return self._layer_rotation_angle

    @layer_rotation_angle.setter
    def layer_rotation_angle(self, value: float):
        self.__validate_range(
            value,
            MachineConstants.MIN_LAYER_ROTATION_ANGLE,
            MachineConstants.MAX_LAYER_ROTATION_ANGLE,
            "layer_rotation_angle",
        )
        self._layer_rotation_angle = value

    @property
    def hatch_spacing(self) -> float:
        """Distance (m) between adjacent scan vectors, or hatches, when
        rastering back and forth with the laser.

        Hatch spacing should allow for a slight overlap of scan vector tracks such that
        some of the material re-melts to ensure full coverage of solid material.

        Valid values are from 6e-5 to 2e-4 m (0.06 and 0.2 mm).
        """
        return self._hatch_spacing

    @hatch_spacing.setter
    def hatch_spacing(self, value: float):
        self.__validate_range(
            value,
            MachineConstants.MIN_HATCH_SPACING,
            MachineConstants.MAX_HATCH_SPACING,
            "hatch_spacing",
        )
        self._hatch_spacing = value

    @property
    def slicing_stripe_width(self) -> float:
        """Width (m) of a stripe, which is a section of scan lines within a
        layer.

        Valid values are from 0.001 to 0.1 m (1 and 100 mm).
        """
        return self._slicing_stripe_width

    @slicing_stripe_width.setter
    def slicing_stripe_width(self, value: float):
        self.__validate_range(
            value,
            MachineConstants.MIN_SLICING_STRIPE_WIDTH,
            MachineConstants.MAX_SLICING_STRIPE_WIDTH,
            "slicing_stripe_width",
        )
        self._slicing_stripe_width = value

    @staticmethod
    def _from_machine_message(msg: MachineMessage):
        """Create an additive machine from a machine message received from the
        Additive service."""
        if isinstance(msg, MachineMessage):
            return AdditiveMachine(
                laser_power=msg.laser_power,
                scan_speed=msg.scan_speed,
                heater_temperature=conversions.kelvin_to_celsius(msg.heater_temperature),
                layer_thickness=msg.layer_thickness,
                beam_diameter=msg.beam_diameter,
                starting_layer_angle=math.degrees(msg.starting_layer_angle),
                layer_rotation_angle=math.degrees(msg.layer_rotation_angle),
                hatch_spacing=msg.hatch_spacing,
                slicing_stripe_width=msg.slicing_stripe_width,
            )
        else:
            raise ValueError("Invalid message type passed to _from_machine_message()")

    def _to_machine_message(self) -> MachineMessage:
        """Create a machine message from the additive machine to send to the
        Additive service."""
        return MachineMessage(
            laser_power=self.laser_power,
            scan_speed=self.scan_speed,
            heater_temperature=conversions.celsius_to_kelvin(self.heater_temperature),
            layer_thickness=self.layer_thickness,
            beam_diameter=self.beam_diameter,
            starting_layer_angle=math.radians(self.starting_layer_angle),
            layer_rotation_angle=math.radians(self.layer_rotation_angle),
            hatch_spacing=self.hatch_spacing,
            slicing_stripe_width=self.slicing_stripe_width,
        )
